""".

Created on Sat Mar 14 16:36:01 2020

@author: mofarrag
"""
import datetime as dt
import math
import os
from collections import OrderedDict
from typing import List, Union

import matplotlib as mpl
import matplotlib.colors as colors
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib import animation, gridspec
from matplotlib.animation import FuncAnimation
from scipy.stats import gumbel_r

hours = list(range(1, 25))


class Visualize:
    """Visualize.

    Visualize class contains different method to animate water surface profile
    spatial data change over time, styling methods for plotting

    Methods
    -------
        1- AnimateArray
        2- CrossSections
        3- WaterSurfaceProfile
        4- CrossSections
    """

    FigureDefaultOptions = dict(
        ylabel="",
        xlabel="",
        legend="",
        legend_size=10,
        figsize=(10, 8),
        labelsize=10,
        fontsize=10,
        name="hist.tif",
        color1="#3D59AB",
        color2="#DC143C",
        linewidth=3,
        Axisfontsize=15,
    )

    linestyles = OrderedDict(
        [
            ("solid", (0, ())),  # 0
            ("loosely dotted", (0, (1, 10))),  # 1
            ("dotted", (0, (1, 5))),  # 2
            ("densely dotted", (0, (1, 1))),  # 3
            ("loosely dashed", (0, (5, 10))),  # 4
            ("dashed", (0, (5, 5))),  # 5
            ("densely dashed", (0, (5, 1))),  # 6
            ("loosely dashdotted", (0, (3, 10, 1, 10))),  # 7
            ("dashdotted", (0, (3, 5, 1, 5))),  # 8
            ("densely dashdotted", (0, (3, 1, 1, 1))),  # 9
            ("loosely dashdotdotted", (0, (3, 10, 1, 10, 1, 10))),  # 10
            ("dashdotdotted", (0, (3, 5, 1, 5, 1, 5))),  # 11
            ("densely dashdotdotted", (0, (3, 1, 1, 1, 1, 1))),  # 12
            ("densely dashdotdottededited", (0, (6, 1, 1, 1, 1, 1))),  # 13
        ]
    )

    MarkerStyleList = [
        "--o",
        ":D",
        "-.H",
        "--x",
        ":v",
        "--|",
        "-+",
        "-^",
        "--s",
        "-.*",
        "-.h",
    ]

    def __init__(self, resolution: str = "Hourly"):
        self.resolution = resolution
        self.Anim = None

    @staticmethod
    def LineStyle(Style: Union[str, int] = "loosely dotted"):
        """LineStyle.

        Line styles for plotting

        Parameters
        ----------
        Style : TYPE, optional
            DESCRIPTION. The default is 'loosely dotted'.

        Returns
        -------
        TYPE
            DESCRIPTION.
        """
        if isinstance(Style, str):
            try:
                return Visualize.linestyles[Style]
            except KeyError:
                msg = (
                    " The Style name you entered-{0}-does not exist please"
                    "choose from the available styles"
                ).format(Style)
                print(msg)
                print(list(Visualize.linestyles))
        else:
            return list(Visualize.linestyles.items())[Style][1]

    @staticmethod
    def MarkerStyle(Style: int):
        """MarkerStyle.

        Marker styles for plotting

        Parameters
        ----------
        Style : TYPE
            DESCRIPTION.

        Returns
        -------
        TYPE
            DESCRIPTION.
        """
        if Style > len(Visualize.MarkerStyleList) - 1:
            Style = Style % len(Visualize.MarkerStyleList)
        return Visualize.MarkerStyleList[Style]

    def GroundSurface(
        self,
        Sub,
        fromxs: Union[str, int] = "",
        toxs: Union[str, int] = "",
        floodplain: bool = False,
        plotlateral: bool = False,
        nxlabels: int = 10,
        figsize: tuple = (20, 10),
        LateralsColor: Union[str, tuple] = "red",
        LaterlasLineWidth: int = 1,
        option: int = 1,
        size: int = 50,
    ):
        """Plot the longitudinal profile of the segment.

        Parameters
        ----------
        Sub : TYPE
            DESCRIPTION.
        fromxs : TYPE, optional
            DESCRIPTION. The default is ''.
        toxs : TYPE, optional
            DESCRIPTION. The default is ''.
        floodplain : TYPE, optional
            DESCRIPTION. The default is False.
        plotlateral : TYPE, optional
            DESCRIPTION. The default is False.
        nxlabels: [int]
            Default is 10
        figsize: [tuple]
            Default is (20, 10)
        LateralsColor: [str, tuple]
            Defaut is "red",
        LaterlasLineWidth: [int]
            Default is 1.
        option: [int]
            Default is 1
        size: [int]
        Default is 50.

        Returns
        -------
        None.
        """
        GroundSurfacefig = plt.figure(70, figsize=figsize)
        gs = gridspec.GridSpec(nrows=2, ncols=6, figure=GroundSurfacefig)
        axGS = GroundSurfacefig.add_subplot(gs[0:2, 0:6])

        if fromxs == "":
            fromxs = Sub.xsname[0]

        if toxs == "":
            toxs = Sub.xsname[-1]

        # not the whole sub-basin
        axGS.set_xticks(list(range(fromxs, toxs)))
        axGS.set_xticklabels(list(range(fromxs, toxs)))

        axGS.set_xlim(fromxs - 1, toxs + 1)

        axGS.tick_params(labelsize=8)
        # plot dikes
        axGS.plot(
            Sub.xsname,
            Sub.crosssections["zl"],
            "k--",
            dashes=(5, 1),
            linewidth=2,
            label="Left Dike",
        )
        axGS.plot(
            Sub.xsname, Sub.crosssections["zr"], "k.-", linewidth=2, label="Right Dike"
        )

        if floodplain:
            fpl = (
                Sub.crosssections["gl"]
                + Sub.crosssections["dbf"]
                + Sub.crosssections["hl"]
            )
            fpr = (
                Sub.crosssections["gl"]
                + Sub.crosssections["dbf"]
                + Sub.crosssections["hr"]
            )
            axGS.plot(Sub.xsname, fpl, "b-.", linewidth=2, label="Floodplain left")
            axGS.plot(Sub.xsname, fpr, "r-.", linewidth=2, label="Floodplain right")

        if plotlateral:
            if isinstance(Sub.LateralsTable, list) and len(Sub.LateralsTable) > 0:
                if option == 1:
                    # plot location of laterals
                    for i in range(len(Sub.LateralsTable)):
                        axGS.vlines(
                            Sub.LateralsTable[i],
                            0,
                            int(Sub.Result1D["q"].max()),
                            colors=LateralsColor,
                            linestyles="dashed",
                            linewidth=LaterlasLineWidth,
                        )
                else:
                    lat = pd.DataFrame()
                    lat["xsid"] = Sub.LateralsTable
                    lat = lat.merge(Sub.crosssections, on="xsid", how="left")

                    axGS.scatter(
                        Sub.LateralsTable,
                        lat["gl"].tolist(),
                        c=LateralsColor,
                        linewidth=LaterlasLineWidth,
                        zorder=10,
                        s=size,
                    )
            else:
                print(" Please Read the Laterals data")

        maxelevel1 = max(
            Sub.crosssections.loc[Sub.crosssections["xsid"] >= fromxs, "zr"][
                Sub.crosssections["xsid"] <= toxs
            ]
        )
        maxelevel2 = max(
            Sub.crosssections.loc[Sub.crosssections["xsid"] >= fromxs, "zl"][
                Sub.crosssections["xsid"] <= toxs
            ]
        )
        maxlelv = max(maxelevel1, maxelevel2)
        minlev = Sub.crosssections.loc[Sub.crosssections["xsid"] == toxs, "gl"].values
        axGS.set_ylim(minlev - 5, maxlelv + 5)

        # plot the bedlevel/baklevel
        if Sub.version == 1:
            axGS.plot(
                Sub.xsname,
                Sub.crosssections["gl"],
                "k-",
                linewidth=5,
                label="Bankful level",
            )
        else:
            axGS.plot(
                Sub.xsname,
                Sub.crosssections["gl"],
                "k-",
                linewidth=5,
                label="Ground level",
            )
            axGS.plot(
                Sub.xsname,
                Sub.crosssections["gl"] + Sub.crosssections["dbf"],
                "k",
                linewidth=2,
                label="Bankful depth",
            )

        if nxlabels != "":
            start, end = axGS.get_xlim()
            label_list = [int(i) for i in np.linspace(start, end, nxlabels)]
            axGS.xaxis.set_ticks(label_list)

        title = "Water surface Profile Simulation Subid = {0}".format(Sub.id)
        axGS.set_title(title, fontsize=15)
        axGS.legend(fontsize=15)
        axGS.set_xlabel("Profile", fontsize=15)
        axGS.set_ylabel("Elevation m", fontsize=15)
        axGS.grid()
        GroundSurfacefig.tight_layout()

    def WaterSurfaceProfile(
        self,
        Sub,
        start: Union[str, dt.datetime],
        end: Union[str, dt.datetime],
        fps: int = 100,
        fromxs: Union[str, int] = "",
        toxs: Union[str, int] = "",
        fmt: str = "%Y-%m-%d",
        figsize: tuple = (20, 10),
        textlocation: tuple = (1, 1),
        LateralsColor: Union[int, str] = "#3D59AB",
        LaterlasLineWidth: int = 1,
        xaxislabelsize: int = 10,
        yaxislabelsize: int = 10,
        nxlabels: int = 10,
        xticklabelsize: int = 8,
        Lastsegment: bool = True,
        floodplain: bool = True,
        repeat: bool = True,
    ) -> FuncAnimation:
        """WaterSurfaceProfile.

        Plot water surface profile

        Parameters
        ----------
        Sub : [Object]
            Sub-object created as a sub class from River object.
        start : [datetime object]
            starting date of the simulation.
        end : [datetime object]
            end date of the simulation.
        fps : [integer], optional
             It is an optional integer value that represents the delay between
             each frame in milliseconds. Its default is 100.
        fromxs : [integer], optional
            number of cross sections to be displayed before the chosen cross
            section . The default is 10.
        toxs : [integer], optional
            number of cross sections to be displayed after the chosen cross
            section . The default is 10.
        xticklabelsize: []

        nxlabels:[]

        yaxislabelsize: []

        LaterlasLineWidth: []

        xaxislabelsize:[]

        LateralsColor: []

        textlocation: []

        fmt: []

        figsize: []

        Lastsegment: [bool]
            Default is True.
        floodplain: [bool]
            Default is True.
        repeat: [bool]
            Defaut is True
        """
        if isinstance(start, str):
            start = dt.datetime.strptime(start, fmt)

        if isinstance(end, str):
            end = dt.datetime.strptime(end, fmt)
        msg = """The start date does not exist in the results, Results are
            between {0} and  {1}""".format(
            Sub.firstday, Sub.lastday
        )
        assert start in Sub.referenceindex_results, msg

        msg = """ The end date does not exist in the results, Results are
            between {0} and  {1}""".format(
            Sub.firstday, Sub.lastday
        )
        assert end in Sub.referenceindex_results, msg

        msg = """please read the boundary condition files using the
            'ReadBoundaryConditions' method """
        assert hasattr(Sub, "QBC"), msg

        msg = """ start Simulation date should be before the end simulation
            date """
        assert start < end, msg

        if Sub.from_beginning == 1:
            Period = Sub.daylist[
                np.where(Sub.referenceindex == start)[0][0] : np.where(
                    Sub.referenceindex == end
                )[0][0]
                + 1
            ]
        else:
            ii = Sub.DateToIndex(start)
            ii2 = Sub.DateToIndex(end)
            Period = list(range(ii, ii2 + 1))

        counter = [(i, j) for i in Period for j in hours]

        fig = plt.figure(60, figsize=figsize)
        gs = gridspec.GridSpec(nrows=2, ncols=6, figure=fig)
        ax1 = fig.add_subplot(gs[0, 2:6])
        ax1.set_ylim(0, int(Sub.Result1D["q"].max()))

        if fromxs == "":
            # xs = 0
            # plot the whole sub-basin
            fromxs = Sub.xsname[0]
        else:
            # xs = 1
            # not the whole sub-basin
            if fromxs < Sub.xsname[0]:
                fromxs = Sub.xsname[0]

        if toxs == "":
            toxs = Sub.xsname[-1]
        else:
            if toxs > Sub.xsname[-1]:
                toxs = Sub.xsname[-1]

        ax1.set_xlim(fromxs - 1, toxs + 1)
        ax1.set_xticks(list(range(fromxs, toxs + 1)))
        ax1.set_xticklabels(list(range(fromxs, toxs + 1)))

        ax1.tick_params(labelsize=xticklabelsize)
        ax1.locator_params(axis="x", nbins=nxlabels)

        ax1.set_xlabel("Cross section No", fontsize=xaxislabelsize)
        ax1.set_ylabel("Discharge (m3/s)", fontsize=yaxislabelsize, labelpad=0.3)
        ax1.set_title("Sub-Basin" + " " + str(Sub.id), fontsize=15)
        ax1.legend(["Discharge"], fontsize=15)

        # plot location of laterals
        for i in range(len(Sub.LateralsTable)):
            ax1.vlines(
                Sub.LateralsTable[i],
                0,
                int(Sub.Result1D["q"].max()),
                colors=LateralsColor,
                linestyles="dashed",
                linewidth=LaterlasLineWidth,
            )

        lat = pd.DataFrame()
        lat["xsid"] = Sub.LateralsTable
        lat = lat.merge(Sub.crosssections, on="xsid", how="left")

        lim = ax1.get_ylim()
        y = np.ones(len(Sub.LateralsTable), dtype=int) * (lim[1] - 50)
        lat = ax1.scatter(
            Sub.LateralsTable,
            y,
            c=LateralsColor,
            linewidth=LaterlasLineWidth,
            zorder=10,
            s=50,
        )

        (q_line,) = ax1.plot([], [], linewidth=5)
        ax1.grid()

        ### BC
        # Q
        ax2 = fig.add_subplot(gs[0, 1])
        ax2.set_xlim(1, 25)
        ax2.set_ylim(0, int(Sub.QBC.max().max()) + 1)

        ax2.set_xlabel("Time", fontsize=yaxislabelsize)
        ax2.set_ylabel("Q (m3/s)", fontsize=yaxislabelsize, labelpad=0.1)
        ax2.set_title("BC - Q", fontsize=20)
        ax2.legend(["Q"], fontsize=15)

        (bc_q_line,) = ax2.plot([], [], linewidth=5)
        bc_q_point = ax2.scatter([], [], s=300)
        ax2.grid()

        # h
        ax3 = fig.add_subplot(gs[0, 0])
        ax3.set_xlim(1, 25)
        ax3.set_ylim(float(Sub.HBC.min().min()), float(Sub.HBC.max().max()))

        ax3.set_xlabel("Time", fontsize=yaxislabelsize)
        ax3.set_ylabel("water level", fontsize=yaxislabelsize, labelpad=0.5)
        ax3.set_title("BC - H", fontsize=20)
        ax3.legend(["WL"], fontsize=10)

        (bc_h_line,) = ax3.plot([], [], linewidth=5)
        bc_h_point = ax3.scatter([], [], s=300)
        ax3.grid()

        # water surface profile
        ax4 = fig.add_subplot(gs[1, 0:6])

        ymax1 = max(
            Sub.crosssections.loc[Sub.crosssections["xsid"] >= fromxs, "zr"][
                Sub.crosssections["xsid"] <= toxs
            ]
        )
        ymax2 = max(
            Sub.crosssections.loc[Sub.crosssections["xsid"] >= fromxs, "zl"][
                Sub.crosssections["xsid"] <= toxs
            ]
        )
        ymax = max(ymax1, ymax2)
        minlev = Sub.crosssections.loc[Sub.crosssections["xsid"] == toxs, "gl"].values
        ax4.set_ylim(minlev - 5, ymax + 5)
        ax4.set_xlim(fromxs - 1, toxs + 1)
        ax4.set_xticks(list(range(fromxs, toxs + 1)))
        ax4.set_xticklabels(list(range(fromxs, toxs + 1)))

        ax4.tick_params(labelsize=xticklabelsize)
        ax4.locator_params(axis="x", nbins=nxlabels)

        ax4.plot(
            Sub.xsname,
            Sub.crosssections["zl"],
            "k--",
            dashes=(5, 1),
            linewidth=2,
            label="Left Dike",
        )
        ax4.plot(
            Sub.xsname, Sub.crosssections["zr"], "k.-", linewidth=2, label="Right Dike"
        )

        if Sub.version == 1:
            ax4.plot(
                Sub.xsname,
                Sub.crosssections["gl"],
                "k-",
                linewidth=5,
                label="Bankful level",
            )
        else:
            ax4.plot(
                Sub.xsname,
                Sub.crosssections["gl"],
                "k-",
                linewidth=5,
                label="Ground level",
            )
            ax4.plot(
                Sub.xsname,
                Sub.crosssections["gl"] + Sub.crosssections["dbf"],
                "k",
                linewidth=2,
                label="Bankful depth",
            )

        if floodplain:
            fpl = (
                Sub.crosssections["gl"]
                + Sub.crosssections["dbf"]
                + Sub.crosssections["hl"]
            )
            fpr = (
                Sub.crosssections["gl"]
                + Sub.crosssections["dbf"]
                + Sub.crosssections["hr"]
            )
            ax4.plot(Sub.xsname, fpl, "b-.", linewidth=2, label="Floodplain left")
            ax4.plot(Sub.xsname, fpr, "r-.", linewidth=2, label="Floodplain right")

        ax4.set_title("Water surface Profile Simulation", fontsize=15)
        ax4.legend(fontsize=15)
        ax4.set_xlabel("Profile", fontsize=yaxislabelsize)
        ax4.set_ylabel("Elevation m", fontsize=yaxislabelsize)
        ax4.grid()

        # plot location of laterals
        for i in range(len(Sub.LateralsTable)):
            ymin = Sub.crosssections.loc[
                Sub.crosssections["xsid"] == Sub.LateralsTable[i], "gl"
            ].values[0]
            ax4.vlines(
                Sub.LateralsTable[i],
                ymin,
                ymax,
                colors=LateralsColor,
                linestyles="dashed",
                linewidth=LaterlasLineWidth,
            )

        day_text = ax4.annotate(
            "",
            xy=(
                fromxs + textlocation[0],
                Sub.crosssections.loc[Sub.crosssections["xsid"] == toxs, "gl"].values
                + textlocation[1],
            ),
            fontsize=20,
        )

        (wl_line,) = ax4.plot([], [], linewidth=5)
        (hLline,) = ax4.plot([], [], linewidth=5)

        gs.update(wspace=0.2, hspace=0.2, top=0.96, bottom=0.1, left=0.05, right=0.96)
        # animation
        plt.show()

        def init_q():
            q_line.set_data([], [])
            wl_line.set_data([], [])
            hLline.set_data([], [])
            day_text.set_text("")

            bc_q_line.set_data([], [])
            bc_h_line.set_data([], [])
            # bc_q_point
            # bc_h_point
            lat.set_sizes([])

            return (
                q_line,
                wl_line,
                hLline,
                day_text,
                bc_q_line,
                bc_h_line,
                bc_q_point,
                bc_h_point,
                lat,
            )

        # animation function. this is called sequentially
        def animate_q(i):
            x = Sub.xsname
            y = Sub.Result1D.loc[Sub.Result1D["day"] == counter[i][0], "q"][
                Sub.Result1D["hour"] == counter[i][1]
            ]
            # the Saintvenant subroutine writes the
            # results of the last xs in the next segment with the current
            # segment
            if not Lastsegment:
                y = y.values[:-1]

            q_line.set_data(x, y)

            day = Sub.referenceindex.loc[counter[i][0], "date"]

            if len(Sub.LateralsTable) > 0:
                lat.set_sizes(
                    sizes=Sub.Laterals.loc[day, Sub.LateralsTable].values * 100
                )

            day_text.set_text("day = " + str(day + dt.timedelta(hours=counter[i][1])))

            y = Sub.Result1D.loc[Sub.Result1D["day"] == counter[i][0], "wl"][
                Sub.Result1D["hour"] == counter[i][1]
            ]
            # the Saintvenant subroutine writes the results
            # of the last xs in the next segment with the current segment
            if not Lastsegment:
                y = y.values[:-1]

            wl_line.set_data(x, y)

            y = (
                Sub.Result1D.loc[Sub.Result1D["day"] == counter[i][0], "h"][
                    Sub.Result1D["hour"] == counter[i][1]
                ]
                * 2
            )
            # temporary as now the Saintvenant subroutine writes the results
            # of the last xs in the next segment with the current segment
            if not Lastsegment:
                y = y.values[:-1]

            y = (
                y
                + Sub.crosssections.loc[
                    Sub.crosssections.index[len(Sub.xsname) - 1], "gl"
                ]
            )
            hLline.set_data(x, y)

            x = Sub.QBC.columns.values

            y = Sub.QBC.loc[Sub.referenceindex.loc[counter[i][0], "date"]].values
            bc_q_line.set_data(x, y)

            y = Sub.HBC.loc[Sub.referenceindex.loc[counter[i][0], "date"]].values
            bc_h_line.set_data(x, y)

            x = counter[i][1]
            y = Sub.referenceindex.loc[counter[i][0], "date"]
            scatter1 = ax2.scatter(x, Sub.QBC[x][y], s=300)
            scatter2 = ax3.scatter(x, Sub.HBC[x][y], s=300)

            return (
                q_line,
                wl_line,
                hLline,
                day_text,
                bc_q_line,
                bc_h_line,
                scatter1,
                scatter1,
                scatter2,
                lat,
            )

        Anim = FuncAnimation(
            fig,
            animate_q,
            init_func=init_q,
            frames=np.shape(counter)[0],
            interval=fps,
            blit=True,
            repeat=repeat,
        )
        self.Anim = Anim
        return Anim

    def WaterSurfaceProfile1Min(
        self,
        Sub,
        start: Union[str, dt.datetime],
        end: Union[str, dt.datetime],
        interval: float = 0.00002,
        fromxs: Union[str, int] = "",
        toxs: Union[str, int] = "",
        fmt: str = "%Y-%m-%d",
        figsize: tuple = (20, 10),
        textlocation: tuple = (1, 1),
        LateralsColor: Union[str, tuple] = "#3D59AB",
        LaterlasLineWidth: int = 1,
        xaxislabelsize: int = 10,
        yaxislabelsize: int = 10,
        nxlabels: int = 20,
        xticklabelsize: int = 8,
        floodplain: bool = True,
        repeat: bool = True,
    ) -> FuncAnimation:
        """WaterSurfaceProfile1Min.

        Plot water surface profile for 1 min data

        Parameters
        ----------
        Sub : TYPE
            DESCRIPTION.
        start : TYPE
            DESCRIPTION.
        end : TYPE
            DESCRIPTION.
        interval : TYPE, optional
            DESCRIPTION. The default is 0.00002.
        fromxs : TYPE, optional
            DESCRIPTION. The default is 10.
        toxs : TYPE, optional
            DESCRIPTION. The default is 10.
        floodplain: [bool]
            Default is True.
        fmt: [str]
            Default is "%Y-%m-%d".
        figsize: [tuple]
            Default is (20, 10).
        textlocation: [tuple]
            Default is (1, 1).
        LateralsColor: [str]
            Default is "#3D59AB".
        LaterlasLineWidth: [int]
            Default is 1.
        xaxislabelsize: [int]
            Default is 10.
        yaxislabelsize: [int]
            Default is 10.
        nxlabels: [int]
            Default is 20.
        xticklabelsize: [int]
            Default is 8.
        floodplain: bool
            Default is True.
        repeat: [bool]
            Defaut is True

        Returns
        -------
        TYPE
            DESCRIPTION.
        """
        start = dt.datetime.strptime(start, fmt)
        end = dt.datetime.strptime(end, fmt) - dt.timedelta(minutes=int(Sub.dt / 60))

        assert start in Sub.h.index, (
            "plot start date in not in the 1min results, the results starts from "
            + str(Sub.h.index[0])
            + " - and ends on "
            + str(Sub.h.index[-1])
        )
        assert end in Sub.h.index, (
            "plot end date in not in the 1min results, the results starts from "
            + str(Sub.h.index[0])
            + " - and ends on "
            + str(Sub.h.index[-1])
        )

        counter = Sub.h.index[
            np.where(Sub.h.index == start)[0][0] : np.where(Sub.h.index == end)[0][0]
        ]
        nstep = (
            len(pd.date_range(start, start + dt.timedelta(days=1), freq=Sub.freq)) - 1
        )

        fig2 = plt.figure(20, figsize=(20, 10))
        gs = gridspec.GridSpec(nrows=2, ncols=6, figure=fig2)

        ax1 = fig2.add_subplot(gs[0, 2:6])

        if fromxs == "":
            fromxs = Sub.xsname[0]
            # toxs = Sub.xsname[-1]
        else:
            if fromxs < Sub.xsname[0]:
                fromxs = Sub.xsname[0]

            # if toxs > Sub.xsname[-1]:
            # toxs = Sub.xsname[-1]

        if toxs == "":
            toxs = Sub.xsname[-1]
        else:
            if toxs > Sub.xsname[-1]:
                toxs = Sub.xsname[-1]

        ax1.set_xlim(fromxs - 1, toxs + 1)

        ax1.set_xticks(list(range(fromxs, toxs + 1)))
        ax1.set_xticklabels(list(range(fromxs, toxs + 1)))

        ax1.tick_params(labelsize=xticklabelsize)
        ax1.locator_params(axis="x", nbins=nxlabels)

        ax1.set_xlabel("Cross section No", fontsize=xaxislabelsize)
        ax1.set_ylabel("Discharge (m3/s)", fontsize=yaxislabelsize, labelpad=0.5)
        ax1.set_title("Sub-Basin" + " " + str(Sub.id), fontsize=15)
        ax1.legend(["Discharge"], fontsize=15)
        ax1.set_ylim(0, int(Sub.q.max().max()))

        if Sub.version < 4:
            # ax1.set_ylim(0, int(Sub.Result1D['q'].max()))

            # plot location of laterals
            for i in range(len(Sub.LateralsTable)):
                ax1.vlines(
                    Sub.LateralsTable[i],
                    0,
                    int(int(Sub.q.max().max())),
                    colors=LateralsColor,
                    linestyles="dashed",
                    linewidth=LaterlasLineWidth,
                )

            lat = pd.DataFrame()
            lat["xsid"] = Sub.LateralsTable
            lat = lat.merge(Sub.crosssections, on="xsid", how="left")

            lim = ax1.get_ylim()
            y = np.ones(len(Sub.LateralsTable), dtype=int) * (lim[1] - 50)
            lat = ax1.scatter(
                Sub.LateralsTable,
                y,
                c=LateralsColor,
                linewidth=LaterlasLineWidth,
                zorder=10,
                s=50,
            )
        else:
            ax1.set_ylim(0, int(Sub.q.max().max()))

        (q_line,) = ax1.plot([], [], linewidth=5)
        ax1.grid()

        ### BC
        # Q
        ax2 = fig2.add_subplot(gs[0, 1:2])
        ax2.set_xlim(1, nstep)
        if Sub.version < 4:
            ax2.set_ylim(0, int(Sub.QBCmin.max().max()))
        else:
            ax2.set_ylim(0, int(Sub.USBC.max()))

        ax2.set_xlabel("Time", fontsize=yaxislabelsize)
        ax2.set_ylabel("Q (m3/s)", fontsize=yaxislabelsize, labelpad=0.1)
        ax2.set_title("BC - Q", fontsize=20)
        ax2.legend(["Q"], fontsize=15)

        (bc_q_line,) = ax2.plot([], [], linewidth=5)
        bc_q_point = ax2.scatter([], [], s=150)
        ax2.grid()

        # h
        ax3 = fig2.add_subplot(gs[0, 0:1])
        ax3.set_xlim(1, nstep)
        if Sub.version < 4:
            ax3.set_ylim(float(Sub.HBCmin.min().min()), float(Sub.HBCmin.max().max()))

        ax3.set_xlabel("Time", fontsize=yaxislabelsize)
        ax3.set_ylabel("water level", fontsize=yaxislabelsize, labelpad=0.5)
        ax3.set_title("BC - H", fontsize=20)
        ax3.legend(["WL"], fontsize=10)

        (bc_h_line,) = ax3.plot([], [], linewidth=5)
        bc_h_point = ax3.scatter([], [], s=150)

        ax3.grid()

        # water surface profile
        ax4 = fig2.add_subplot(gs[1, 0:6])

        ymax1 = max(
            Sub.crosssections.loc[Sub.crosssections["xsid"] >= fromxs, "zr"][
                Sub.crosssections["xsid"] <= toxs
            ]
        )
        ymax2 = max(
            Sub.crosssections.loc[Sub.crosssections["xsid"] >= fromxs, "zl"][
                Sub.crosssections["xsid"] <= toxs
            ]
        )
        ymax = max(ymax1, ymax2)
        minlev = Sub.crosssections.loc[Sub.crosssections["xsid"] == toxs, "gl"].values
        ax4.set_ylim(minlev - 5, ymax + 5)
        ax4.set_xlim(fromxs - 1, toxs + 1)
        ax4.set_xticks(list(range(fromxs, toxs + 1)))

        ax4.tick_params(labelsize=xaxislabelsize)
        ax4.locator_params(axis="x", nbins=nxlabels)

        ax4.plot(
            Sub.xsname,
            Sub.crosssections["zl"],
            "k--",
            dashes=(5, 1),
            linewidth=2,
            label="Left Dike",
        )
        ax4.plot(
            Sub.xsname, Sub.crosssections["zr"], "k.-", linewidth=2, label="Right Dike"
        )

        if Sub.version == 1:
            ax4.plot(
                Sub.xsname,
                Sub.crosssections["gl"],
                "k-",
                linewidth=5,
                label="Bankful level",
            )
        else:
            ax4.plot(
                Sub.xsname,
                Sub.crosssections["gl"],
                "k-",
                linewidth=5,
                label="Ground level",
            )
            ax4.plot(
                Sub.xsname,
                Sub.crosssections["gl"] + Sub.crosssections["dbf"],
                "k",
                linewidth=2,
                label="Bankful depth",
            )

        if floodplain:
            fpl = (
                Sub.crosssections["gl"]
                + Sub.crosssections["dbf"]
                + Sub.crosssections["hl"]
            )
            fpr = (
                Sub.crosssections["gl"]
                + Sub.crosssections["dbf"]
                + Sub.crosssections["hr"]
            )
            ax4.plot(Sub.xsname, fpl, "b-.", linewidth=2, label="Floodplain left")
            ax4.plot(Sub.xsname, fpr, "r-.", linewidth=2, label="Floodplain right")

        ax4.set_title("Water surface Profile Simulation", fontsize=15)
        ax4.legend(fontsize=10)
        ax4.set_xlabel("Profile", fontsize=10)
        ax4.set_ylabel("Elevation m", fontsize=10)
        ax4.grid()

        # plot location of laterals
        for i in range(len(Sub.LateralsTable)):
            ymin = Sub.crosssections.loc[
                Sub.crosssections["xsid"] == Sub.LateralsTable[i], "gl"
            ].values[0]
            ax4.vlines(
                Sub.LateralsTable[i],
                ymin,
                ymax,
                colors=LateralsColor,
                linestyles="dashed",
                linewidth=LaterlasLineWidth,
            )

        day_text = ax4.annotate(
            "",
            xy=(
                fromxs + textlocation[0],
                Sub.crosssections.loc[Sub.crosssections["xsid"] == toxs, "gl"].values
                + textlocation[1],
            ),
            fontsize=20,
        )

        (wl_line,) = ax4.plot([], [], linewidth=5)
        (hLline,) = ax4.plot([], [], linewidth=5)

        gs.update(wspace=0.2, hspace=0.2, top=0.96, bottom=0.1, left=0.05, right=0.96)
        # animation
        plt.show()

        # animation
        def init_min():
            q_line.set_data([], [])
            wl_line.set_data([], [])
            day_text.set_text("")
            bc_q_line.set_data([], [])
            bc_h_line.set_data([], [])
            # bc_q_point
            # bc_h_point
            lat.set_sizes([])

            return (
                q_line,
                wl_line,
                bc_q_line,
                bc_h_line,
                bc_q_point,
                bc_h_point,
                day_text,
                lat,
            )

        # animation function. this is called sequentially
        def animate_min(i):

            day_text.set_text("Date = " + str(counter[i]))
            # discharge (ax1)
            x = Sub.xsname
            y = Sub.q[Sub.q.index == counter[i]].values[0]
            q_line.set_data(x, y)

            # water level (ax4)
            y = Sub.h.loc[Sub.q.index == counter[i]].values[0]
            wl_line.set_data(x, y)

            day = counter[i].floor(freq="D")

            lat.set_sizes(sizes=Sub.Laterals.loc[day, Sub.LateralsTable].values * 100)

            # BC Q (ax2)

            x = Sub.QBCmin.columns.values

            y = Sub.QBCmin.loc[day].values
            bc_q_line.set_data(x, y)

            # BC H (ax3)
            y = Sub.HBCmin.loc[day].values
            bc_h_line.set_data(x, y)

            # BC Q point (ax2)
            x = ((counter[i] - day).seconds / 60) + 1
            scatter1 = ax2.scatter(x, Sub.QBCmin[x][day], s=150)
            # BC h point (ax3)
            scatter2 = ax3.scatter(x, Sub.HBCmin[x][day], s=150)

            return (
                q_line,
                wl_line,
                bc_q_line,
                bc_h_line,
                scatter1,
                scatter2,
                day_text,
                lat,
            )

        anim = FuncAnimation(
            fig2,
            animate_min,
            init_func=init_min,
            frames=np.shape(counter)[0],
            interval=interval,
            blit=True,
            repeat=repeat,
        )
        self.Anim = anim
        return anim

    def river1d(
        self,
        Sub,
        start,
        end,
        interval=0.00002,
        xs=0,
        xsbefore=10,
        xsafter=10,
        fmt="%Y-%m-%d",
        textlocation=2,
        xaxislabelsize=15,
        yaxislabelsize=15,
        nxlabels=50,
        plotbanhfuldepth=False,
    ):
        """river1d.

        plot river 1D

        Parameters
        ----------
        Sub : TYPE
            DESCRIPTION.
        start : TYPE
            DESCRIPTION.
        end : TYPE
            DESCRIPTION.
        interval : TYPE, optional
            DESCRIPTION. The default is 0.00002.
        xs : TYPE, optional
            DESCRIPTION. The default is 0.
        xsbefore : TYPE, optional
            DESCRIPTION. The default is 10.
        xsafter : TYPE, optional
            DESCRIPTION. The default is 10.

        Returns
        -------
        TYPE
            DESCRIPTION.
        """
        if isinstance(start, str):
            start = dt.datetime.strptime(start, fmt)

        if isinstance(end, str):
            end = dt.datetime.strptime(end, fmt)

        msg = (
            "plot start date in not in the 1min results, the results starts"
            " from {0} - and ends on {1} "
        )

        assert start in Sub.referenceindex_results, msg.format(
            Sub.referenceindex_results[0], Sub.referenceindex_results[-1]
        )

        assert end in Sub.referenceindex_results, msg.format(
            Sub.referenceindex_results[0], Sub.referenceindex_results[-1]
        )

        counter = Sub.referenceindex_results[
            np.where(Sub.referenceindex_results == start)[0][0] : np.where(
                Sub.referenceindex_results == end
            )[0][0]
            + 1
        ]

        margin = 10
        fig2 = plt.figure(20, figsize=(20, 10))
        gs = gridspec.GridSpec(nrows=2, ncols=6, figure=fig2)

        # USBC
        ax1 = fig2.add_subplot(gs[0, 0])
        # ax1.set_xlim(1, 1440)
        ax1.set_ylim(0, int(Sub.usbc.max() + margin))
        ax1.set_xlabel("Time", fontsize=15)
        if Sub.usbc.columns[0] == "q":
            # ax1.set_ylabel('USBC - Q (m3/s)', fontsize=15)
            ax1.set_title("USBC - Q (m3/s)", fontsize=20)
        else:
            # ax1.set_ylabel('USBC - H (m)', fontsize=15)
            ax1.set_title("USBC - H (m)", fontsize=20)
        # ax1.legend(["Q"], fontsize=10)
        ax1.set_xlim(1, 25)
        (usbc_line,) = ax1.plot([], [], linewidth=5)
        # usbc_point = ax1.scatter([], [], s=150)
        ax1.grid()

        ax2 = fig2.add_subplot(gs[0, 1:5])
        if xs == 0:
            # plot the whole sub-basin
            ax2.set_xlim(Sub.xsname[0] - 1, Sub.xsname[-1] + 1)
            ax2.set_xticks(Sub.xsname)
            ax2.set_xticklabels(Sub.xsname)

            FigureFirstXS = Sub.xsname[0]
            FigureLastXS = Sub.xsname[-1]
        else:
            # not the whole sub-basin
            FigureFirstXS = Sub.xsname[xs] - xsbefore
            if FigureFirstXS < Sub.xsname[0]:
                FigureFirstXS = Sub.xsname[0]

            FigureLastXS = Sub.xsname[xs] + xsafter
            if FigureLastXS > Sub.xsname[-1]:
                FigureLastXS = Sub.xsname[-1]

            ax2.set_xlim(FigureFirstXS, FigureLastXS)
            ax2.set_xticks(list(range(FigureFirstXS, FigureLastXS)))
            ax2.set_xticklabels(list(range(FigureFirstXS, FigureLastXS)))

        ax2.set_ylim(np.nanmin(Sub.q) - 10, int(np.nanmax(Sub.q)) + 10)
        ax2.tick_params(labelsize=xaxislabelsize)
        ax2.locator_params(axis="x", nbins=nxlabels)
        ax2.set_xlabel("Cross section No", fontsize=xaxislabelsize)
        ax2.set_title("Discharge (m3/s)", fontsize=20)
        # ax2.set_ylabel('Discharge (m3/s)', fontsize=yaxislabelsize, labelpad=0.5)
        ax2.legend(["Discharge"], fontsize=15)

        (q_line,) = ax2.plot([], [], linewidth=5)
        ax2.grid()

        ### BC

        # DSBC
        ax3 = fig2.add_subplot(gs[0, 5:6])
        ax3.set_xlim(1, 25)
        ax3.set_ylim(0, float(Sub.dsbc.min() + margin))

        ax3.set_xlabel("Time", fontsize=15)
        if Sub.dsbc.columns[0] == "q":
            # ax3.set_ylabel('DSBC', fontsize=15, labelpad=0.5)
            ax3.set_title("DSBC - Q (m3/s)", fontsize=20)
        else:
            # ax3.set_ylabel('USBC', fontsize=15, labelpad=0.5)
            ax3.set_title("DSBC - H(m)", fontsize=20)

        # ax3.legend(["WL"], fontsize=10)

        (dsbc_line,) = ax3.plot([], [], linewidth=5)
        # dsbc_point = ax3.scatter([], [], s=300)
        ax3.grid()

        # water surface profile
        ax4 = fig2.add_subplot(gs[1, 0:6])

        if xs == 0:
            ax4.set_xlim(Sub.xsname[0] - 1, Sub.xsname[-1] + 1)
            ax4.set_xticks(Sub.xsname)
            ymin = Sub.crosssections.loc[
                Sub.crosssections["xsid"] == FigureFirstXS, "bed level"
            ].values.min()
            ymax = Sub.crosssections.loc[
                Sub.crosssections["xsid"] == FigureFirstXS, "bed level"
            ].values.max()
            ax4.set_ylim(ymin, ymax + np.nanmax(Sub.h) + 5)
        else:
            ax4.set_xlim(FigureFirstXS, FigureLastXS)
            ax4.set_xticks(list(range(FigureFirstXS, FigureLastXS)))
            ax4.set_ylim(
                Sub.crosssections.loc[
                    Sub.crosssections["xsid"] == FigureFirstXS, "bed level"
                ].values,
                Sub.crosssections.loc[
                    Sub.crosssections["xsid"] == FigureLastXS, "zr"
                ].values
                + 5,
            )

        ax4.tick_params(labelsize=xaxislabelsize)
        ax4.locator_params(axis="x", nbins=nxlabels)

        ax4.plot(
            Sub.xsname,
            Sub.crosssections["bed level"],
            "k-",
            linewidth=5,
            label="Ground level",
        )
        if plotbanhfuldepth:
            ax4.plot(
                Sub.xsname,
                Sub.crosssections["bed level"] + Sub.crosssections["depth"],
                "k",
                linewidth=2,
                label="Bankful depth",
            )

        ax4.set_title("Water surface Profile Simulation", fontsize=15)
        ax4.legend(["Ground level", "Bankful depth"], fontsize=10)
        ax4.set_xlabel("Profile", fontsize=10)
        ax4.set_ylabel("Elevation m", fontsize=10)
        ax4.grid()

        if xs == 0:
            textlocation = textlocation + Sub.xsname[0]
            day_text = ax4.annotate(
                " ",
                xy=(textlocation, Sub.crosssections["bed level"].min() + 1),
                fontsize=20,
            )
        else:
            day_text = ax4.annotate(
                " ",
                xy=(
                    FigureFirstXS + textlocation,
                    Sub.crosssections.loc[
                        Sub.crosssections["xsid"] == FigureLastXS, "gl"
                    ].values
                    + 1,
                ),
                fontsize=20,
            )

        (wl_line,) = ax4.plot([], [], linewidth=5)

        gs.update(wspace=0.2, hspace=0.2, top=0.96, bottom=0.1, left=0.05, right=0.96)
        # animation
        plt.show()

        # animation
        def init_min():
            q_line.set_data([], [])
            wl_line.set_data([], [])
            day_text.set_text("")
            usbc_line.set_data([], [])
            dsbc_line.set_data([], [])
            # bc_q_point
            return q_line, wl_line, day_text, usbc_line, dsbc_line

        # animation function. this is called sequentially
        def animate_min(i):
            day_text.set_text("Date = " + str(counter[i]))

            # discharge (ax1)
            x = Sub.xsname

            y = Sub.q[np.where(Sub.referenceindex_results == counter[i])[0][0], :]
            q_line.set_data(x, y)

            # water level (ax4)
            y = Sub.wl[np.where(Sub.referenceindex_results == counter[i])[0][0], :]
            wl_line.set_data(x, y)

            # USBC
            f = dt.datetime(counter[i].year, counter[i].month, counter[i].day)
            y = (
                Sub.usbc.loc[f : f + dt.timedelta(days=1)]
                .resample("H")
                .mean()
                .interpolate("linear")
                .values
            )
            x = hours[: len(y)]
            usbc_line.set_data(x, y)

            # DSBC
            y = (
                Sub.dsbc.loc[f : f + dt.timedelta(days=1)]
                .resample("H")
                .mean()
                .interpolate("linear")
                .values
            )
            dsbc_line.set_data(x, y)

            # x = counter[i][1]
            # y = Sub.referenceindex.loc[counter[i][0], 'date']
            # ax2.scatter(x, Sub.QBC[x][y])

            # # BC Q point (ax2)
            # x = ((counter[i] - dt.datetime(counter[i].year, counter[i].month, counter[i].day)).seconds / 60) + 1
            # y = dt.datetime(counter[i].year, counter[i].month, counter[i].day)
            # ax2.scatter(x, Sub.USBC[x][y])

            return (
                q_line,
                wl_line,
                day_text,
                usbc_line,
                dsbc_line,
            )  # ax2.scatter(x, Sub.USBC[x][y], s=150),

        # plt.tight_layout()

        anim = FuncAnimation(
            fig2,
            animate_min,
            init_func=init_min,
            frames=len(counter),
            interval=interval,
            blit=True,
        )

        return anim

    def CrossSections(
        self,
        Sub,
        fromxs: Union[str, int] = "",
        toxs: Union[str, int] = "",
        xsrows: int = 3,
        xscolumns: int = 3,
        bedlevel: bool = False,
        titlesize: int = 15,
        textsize: int = 15,
        figsize: tuple = (18, 10),
        linewidth: int = 6,
        samescale: bool = False,
        textspacing: List[tuple] = [(1, 1), (1, 2)],
        plottingoption: int = 1,
        plotannotation: bool = True,
    ):
        """CrossSections.

        Plot cross sections of a river segment.

        Parameters
        ----------
        Sub : [Object]
            Sub-object created as a sub class from River object..
        fromxs : TYPE, optional
            DESCRIPTION. The default is ''.
        toxs : TYPE, optional
            DESCRIPTION. The default is ''.
        xsrows : TYPE, optional
            DESCRIPTION. The default is 3.
        xscolumns : TYPE, optional
            DESCRIPTION. The default is 3.
        bedlevel : TYPE, optional
            DESCRIPTION. The default is False.
        titlesize : TYPE, optional
            DESCRIPTION. The default is 15.
        textsize : TYPE, optional
            DESCRIPTION. The default is 15.
        figsize : TYPE, optional
            DESCRIPTION. The default is (18, 10).
        linewidth : TYPE, optional
            DESCRIPTION. The default is 6.
        plottingoption : [integer]
            1 if you want to plot the whole cross-section, 2 to execlude the
            dikes(river bankfull area and floodplain will be plotted),
            3 to plot only the bankfull area.
        samescale: [bool]
            Default is False.
        textspacing: [tuple]
            Default is [(1, 1), (1, 2)].
        plotannotation: [bool]
            Default is True.

        Returns
        -------
        None.
        """
        if fromxs == "":
            startxs_ind = 0
        else:
            startxs_ind = Sub.xsname.index(fromxs)

        if toxs == "":
            endxs_ind = Sub.xsno - 1
        else:
            endxs_ind = Sub.xsname.index(toxs)

        names = list(range(1, 17))
        XSS = pd.DataFrame(
            columns=names, index=Sub.crosssections.loc[startxs_ind:endxs_ind, "xsid"]
        )

        # calculate the vertices of the cross sections
        for i in range(startxs_ind, endxs_ind + 1):
            ind = XSS.index[i - startxs_ind]
            ind2 = Sub.crosssections.index[i]

            XSS[1].loc[XSS.index == ind] = 0
            XSS[2].loc[XSS.index == ind] = 0

            bl = Sub.crosssections["bl"].loc[Sub.crosssections.index == ind2].values[0]
            b = Sub.crosssections["b"].loc[Sub.crosssections.index == ind2].values[0]
            br = Sub.crosssections["br"].loc[Sub.crosssections.index == ind2].values[0]

            XSS[3].loc[XSS.index == ind] = bl
            XSS[4].loc[XSS.index == ind] = bl
            XSS[5].loc[XSS.index == ind] = bl + b
            XSS[6].loc[XSS.index == ind] = bl + b
            XSS[7].loc[XSS.index == ind] = bl + b + br
            XSS[8].loc[XSS.index == ind] = bl + b + br

            gl = Sub.crosssections["gl"].loc[Sub.crosssections.index == ind2].values[0]

            if bedlevel:
                subtract = 0
            else:
                subtract = gl

            zl = Sub.crosssections["zl"].loc[Sub.crosssections.index == ind2].values[0]
            zr = Sub.crosssections["zr"].loc[Sub.crosssections.index == ind2].values[0]

            if Sub.version > 1:
                dbf = (
                    Sub.crosssections["dbf"]
                    .loc[Sub.crosssections.index == ind2]
                    .values[0]
                )

            hl = Sub.crosssections["hl"].loc[Sub.crosssections.index == ind2].values[0]
            hr = Sub.crosssections["hr"].loc[Sub.crosssections.index == ind2].values[0]

            XSS[9].loc[XSS.index == ind] = zl - subtract

            if Sub.version == 1:
                XSS[10].loc[XSS.index == ind] = gl + hl - subtract
                XSS[11].loc[XSS.index == ind] = gl - subtract
                XSS[14].loc[XSS.index == ind] = gl - subtract
                XSS[15].loc[XSS.index == ind] = gl + hr - subtract
            else:
                XSS[10].loc[XSS.index == ind] = gl + dbf + hl - subtract
                XSS[11].loc[XSS.index == ind] = gl + dbf - subtract
                XSS[14].loc[XSS.index == ind] = gl + dbf - subtract
                XSS[15].loc[XSS.index == ind] = gl + dbf + hr - subtract

            XSS[12].loc[XSS.index == ind] = gl - subtract
            XSS[13].loc[XSS.index == ind] = gl - subtract

            XSS[16].loc[XSS.index == ind] = zr - subtract

        # plot the cross sections
        xsplot = len(range(startxs_ind, endxs_ind + 1))
        figno = int(math.ceil(xsplot / (xscolumns * xsrows)))

        ind2 = startxs_ind
        ind = XSS.index[ind2 - startxs_ind]
        for i in range(figno):

            # fig = plt.figure(1000 + i, figsize=figsize)
            # -----------------
            if samescale:
                sharex = True
                sharey = True
            else:
                sharex = False
                sharey = False

            fig, ax_XS = plt.subplots(
                ncols=xscolumns,
                nrows=xsrows,
                figsize=figsize,
                sharex=sharex,
                sharey=sharey,
            )
            # gs = gridspec.GridSpec(xsrows, xscolumns)

            for j in range(xsrows):
                for k in range(xscolumns):
                    if ind2 <= endxs_ind:
                        XsId = Sub.crosssections["xsid"][Sub.crosssections.index[ind2]]
                        xcoord = (
                            XSS[names[0:8]].loc[XSS.index == ind].values.tolist()[0]
                        )
                        ycoord = (
                            XSS[names[8:16]].loc[XSS.index == ind].values.tolist()[0]
                        )
                        b = (
                            Sub.crosssections["b"]
                            .loc[Sub.crosssections["xsid"] == ind]
                            .values[0]
                        )
                        bl = (
                            Sub.crosssections["bl"]
                            .loc[Sub.crosssections["xsid"] == ind]
                            .values[0]
                        )
                        gl = (
                            Sub.crosssections["gl"]
                            .loc[Sub.crosssections["xsid"] == ind]
                            .values[0]
                        )

                        # ax_XS = fig.add_subplot(gs[j, k])
                        if plottingoption == 1:
                            ax_XS[j, k].plot(xcoord, ycoord, linewidth=linewidth)
                            x = textspacing[0][0]
                            x1 = textspacing[1][0]
                        elif plottingoption == 2:
                            ax_XS[j, k].plot(
                                xcoord[1:-1], ycoord[1:-1], linewidth=linewidth
                            )
                            x = textspacing[0][0] + bl
                            x1 = textspacing[1][0] + bl
                        else:
                            ax_XS[j, k].plot(
                                xcoord[2:-2], ycoord[2:-2], linewidth=linewidth
                            )
                            x = textspacing[0][0] + bl
                            x1 = textspacing[1][0] + bl

                        ax_XS[j, k].title.set_text("xs ID = " + str(XsId))
                        ax_XS[j, k].title.set_fontsize(titlesize)

                        if samescale:
                            # when sharex and sharey are true the labels
                            # disappear so set thier visability to true
                            ax_XS[j, k].xaxis.set_tick_params(labelbottom=True)
                            ax_XS[j, k].yaxis.set_tick_params(labelbottom=True)

                        if plotannotation:
                            if Sub.version > 1:
                                dbf = (
                                    Sub.crosssections["dbf"]
                                    .loc[Sub.crosssections["xsid"] == ind]
                                    .values[0]
                                )
                                b = (
                                    Sub.crosssections["b"]
                                    .loc[Sub.crosssections["xsid"] == ind]
                                    .values[0]
                                )

                                if bedlevel:
                                    ax_XS[j, k].annotate(
                                        "dbf=" + str(round(dbf, 2)),
                                        xy=(x, gl + textspacing[0][1]),
                                        fontsize=textsize,
                                    )

                                    ax_XS[j, k].annotate(
                                        "b=" + str(round(b, 2)),
                                        xy=(x1, gl + textspacing[1][1]),
                                        fontsize=textsize,
                                    )
                                else:

                                    ax_XS[j, k].annotate(
                                        "dbf=" + str(round(dbf, 2)),
                                        xy=(x, textspacing[0][1]),
                                        fontsize=textsize,
                                    )

                                    ax_XS[j, k].annotate(
                                        "b=" + str(round(b, 2)),
                                        xy=(x1, textspacing[1][1]),
                                        fontsize=textsize,
                                    )

                        ind2 = ind2 + 1
                        ind = ind + 1

            plt.subplots_adjust(
                wspace=0.2, hspace=0.2, top=0.96, bottom=0.1, left=0.05, right=0.96
            )

        return fig, ax_XS

    def Plot1minProfile(
        self,
        Sub,
        date: str,
        xaxislabelsize: int = 10,
        nxlabels: int = 50,
        fmt: str = "%Y-%m-%d",
    ):
        """Plot1minProfile.

        Plot water surface profile for 1 min data.

        Parameters
        ----------
        Sub : TYPE
            DESCRIPTION.
        date : TYPE
            DESCRIPTION.
        xaxislabelsize : TYPE, optional
            DESCRIPTION. The default is 10.
        nxlabels : TYPE, optional
            DESCRIPTION. The default is 50.
        fmt : TYPE, optional
            DESCRIPTION. The default is "%Y-%m-%d".

        Returns
        -------
        None.
        """
        if isinstance(date, str):
            date = dt.datetime.strptime(date, fmt)

        fig50, ax1 = plt.subplots(figsize=(15, 8))
        ax2 = ax1.twinx()
        ax1.plot(Sub.q.columns, Sub.q[Sub.q.index == date].values[0], "r")
        ax2.plot(Sub.h.columns, Sub.h[Sub.q.index == date].values[0])
        ax1.set_ylabel("Discharge", fontsize=20)
        ax2.set_ylabel("Water level", fontsize=20)
        ax1.set_xlabel("Cross sections", fontsize=20)
        ax1.set_xticks(Sub.xsname)
        ax1.tick_params(labelsize=xaxislabelsize)
        ax1.locator_params(axis="x", nbins=nxlabels)

        ax1.grid()

    @staticmethod
    def SaveProfileAnimation(Anim, Path="", fps=3, ffmpegPath=""):
        """SaveProfileAnimation.

        save video animation
        available extentions .mov, .gif, .avi, .mp4

        Parameters
        ----------
        Anim : TYPE
            DESCRIPTION.
        Path : TYPE, optional
            DESCRIPTION. The default is ''.
        fps : TYPE, optional
            DESCRIPTION. The default is 3.
        ffmpegPath : TYPE, optional
            DESCRIPTION. The default is ''.

        in order to save a video using matplotlib you have to download ffmpeg
        from https://ffmpeg.org/ and define this path to matplotlib

        import matplotlib as mpl
        mpl.rcParams['animation.ffmpeg_path'] = "path where you saved the
                                                ffmpeg.exe/ffmpeg.exe"

        Returns
        -------
        None.
        """
        message = """
            please visit https://ffmpeg.org/ and download a version of ffmpeg
            compitable with your operating system, and copy the content of the
            folder and paste it in the "c:/user/.matplotlib/ffmpeg-static/"
            """
        if ffmpegPath == "":
            ffmpegPath = os.getenv("HOME") + "/.matplotlib/ffmpeg-static/bin/ffmpeg.exe"
            assert os.path.exists(ffmpegPath), "{0}".format(message)

        mpl.rcParams["animation.ffmpeg_path"] = ffmpegPath

        Ext = Path.split(".")[1]
        if Ext == "gif":
            msg = """ please enter a valid path to save the animation"""
            assert len(Path) >= 1 and Path.endswith(".gif"), "{0}".format(msg)

            writergif = animation.PillowWriter(fps=fps)
            Anim.save(Path, writer=writergif)
        else:
            try:
                if Ext == "avi" or Ext == "mov":
                    writervideo = animation.FFMpegWriter(fps=fps, bitrate=1800)
                    Anim.save(Path, writer=writervideo)
                elif Ext == "mp4":
                    writermp4 = animation.FFMpegWriter(fps=fps, bitrate=1800)
                    Anim.save(Path, writer=writermp4)

            except FileNotFoundError:
                msg = """ please visit https://ffmpeg.org/ and download a
                    version of ffmpeg compitable with your operating system, for
                    more details please check the method definition
                    """
                print("{0}".format(msg))

    def SaveAnimation(anim, VideoFormat="gif", Path="", SaveFrames=20):
        """saveAnimation.

        Save animation

        Parameters
        ----------
        anim : TYPE
            DESCRIPTION.
        VideoFormat : TYPE, optional
            DESCRIPTION. The default is "gif".
        Path : TYPE, optional
            DESCRIPTION. The default is ''.
        SaveFrames : TYPE, optional
            DESCRIPTION. The default is 20.

        in order to save a video using matplotlib you have to download ffmpeg from
        https://ffmpeg.org/ and define this path to matplotlib

        import matplotlib as mpl
        mpl.rcParams['animation.ffmpeg_path'] = "path where you saved the ffmpeg.exe/ffmpeg.exe"

        Returns
        -------
        None.
        """
        ffmpegPath = os.getenv("HOME") + "/.matplotlib/ffmpeg-static/bin/ffmpeg.exe"

        message = """
        please visit https://ffmpeg.org/ and download a version of ffmpeg compitable
        with your operating system, and copy the content of the folder and paste it
        in the "c:/user/.matplotlib/ffmpeg-static/"
        """
        assert os.path.exists(ffmpegPath), message

        mpl.rcParams["animation.ffmpeg_path"] = ffmpegPath

        if VideoFormat == "gif":
            writergif = animation.PillowWriter(fps=SaveFrames)
            anim.save(Path, writer=writergif)
        else:
            try:
                if VideoFormat == "avi" or VideoFormat == "mov":
                    writervideo = animation.FFMpegWriter(fps=SaveFrames, bitrate=1800)
                    anim.save(Path, writer=writervideo)
                elif VideoFormat == "mp4":
                    writermp4 = animation.FFMpegWriter(fps=SaveFrames, bitrate=1800)
                    anim.save(Path, writer=writermp4)
            except FileNotFoundError:
                print(
                    "please visit https://ffmpeg.org/ and download a version of ffmpeg compitable with your operating system, for more details please check the method definition"
                )

    @staticmethod
    def Histogram(v1, v2, NoAxis=2, filter1=0.2, Save=False, pdf=True, **kwargs):
        """Histogram.

        Histogram method plots the histogram of two given list of values

        Parameters
        ----------
            1-v1 : [List]
                first list of values.
            2-v2 : [List]
                second list of values.

        Returns
        -------
            - histogram plot.

        Example
        -------
        Vis.Histogram(Val1, val2,2,figsize=(5.5,4.5), color1='#27408B',
                        xlabel = 'Inundation Depth (m)', ylabel = 'Frequency', legend_size = 15,
                         fontsize=15, labelsize = 15, Axisfontsize = 11,
                         legend = ['RIM1.0', 'RIM2.0'], pdf = False, Save = False,
                         name = str(Event1.EventIndex.loc[EndInd,'id']))
        """
        # update the default options
        Fkeys = list(kwargs.keys())
        for key in Fkeys:
            if key in Visualize.FigureDefaultOptions.keys():
                Visualize.FigureDefaultOptions[key] = kwargs[key]

        v1 = np.array([j for j in v1 if j > filter1])
        v2 = np.array([j for j in v2 if j > filter1])

        if pdf:
            param_dist1 = gumbel_r.fit(np.array(v1))
            param_dist2 = gumbel_r.fit(np.array(v2))

            d1 = np.linspace(v1.min(), v1.max(), v1.size)
            d2 = np.linspace(v2.min(), v2.max(), v2.size)
            pdf_fitted1 = gumbel_r.pdf(d1, loc=param_dist1[0], scale=param_dist1[1])
            pdf_fitted2 = gumbel_r.pdf(d2, loc=param_dist2[0], scale=param_dist2[1])

        if NoAxis == 1:
            # if bins in kwargs.keys():
            plt.figure(60, figsize=(10, 8))
            n, bins, patches = plt.hist([v1, v2], color=["#3D59AB", "#DC143C"])
            # for key in kwargs.keys():
            #     if key == 'legend':
            #         plt.legend(kwargs['legend'])
            #     if key == 'legend size':
            #         plt.legend(kwargs['legend'],fontsize = int(kwargs['legend_size']))
            #     if key == 'xlabel':
            #         plt.xlabel(kwargs['xlabel'])
            #     if key == 'ylabel':
            #         plt.ylabel(kwargs['ylabel'])
            # #     # if key == 'xlabel':
            # #         # xlabel = kwargs['xlabel']
            # #     # if key == 'xlabel':
            # #         # xlabel = kwargs['xlabel']

        elif NoAxis == 2:
            fig, ax1 = plt.subplots(figsize=Visualize.FigureDefaultOptions["figsize"])
            # n1= ax1.hist([v1,v2], bins=15, alpha = 0.7, color=[color1,color2])
            n1 = ax1.hist(
                [v1, v2],
                bins=15,
                alpha=0.7,
                color=[
                    Visualize.FigureDefaultOptions["color1"],
                    Visualize.FigureDefaultOptions["color2"],
                ],
            )
            # label=['RIM1.0','RIM2.0']) #width = 0.2,

            ax1.set_ylabel(
                "Frequency", fontsize=Visualize.FigureDefaultOptions["AxisLabelSize"]
            )
            # ax1.yaxis.label.set_color(color1)
            ax1.set_xlabel(
                "Inundation Depth Ranges (m)",
                fontsize=Visualize.FigureDefaultOptions["AxisLabelSize"],
            )

            # ax1.tick_params(axis='y', color = color1)
            # ax1.spines['right'].set_color(color1)
            if pdf:
                ax2 = ax1.twinx()
                ax2.plot(
                    d1,
                    pdf_fitted1,
                    "-.",
                    color=Visualize.FigureDefaultOptions["color1"],
                    linewidth=Visualize.FigureDefaultOptions["linewidth"],
                    label="RIM1.0 pdf",
                )
                ax2.plot(
                    d2,
                    pdf_fitted2,
                    "-.",
                    color=Visualize.FigureDefaultOptions["color2"],
                    linewidth=Visualize.FigureDefaultOptions["linewidth"],
                    label="RIM2.0 pdf",
                )
                ax2.set_ylabel(
                    "Probability density function (pdf)",
                    fontsize=Visualize.FigureDefaultOptions["AxisLabelSize"],
                )
            # else:
            #     ax2.yaxis.set_ticklabels([])
            #     # ax2.yaxis.set_major_formatter(plt.NullFormatter())
            #     # ax2.tick_params(right='off', labelright='off')
            #     ax2.set_xticks([])
            #     ax2.tick_params(axis='y', color = color2)

            #     # if key == 'xlabel':
            #         # xlabel = kwargs['xlabel']
            #     # if key == 'xlabel':
            #         # xlabel = kwargs['xlabel']

            # n2 = ax2.hist(v2,  bins=n1[1], alpha = 0.4, color=color2)#width=0.2,
            # ax2.set_ylabel("Frequency", fontsize = 15)
            # ax2.yaxis.label.set_color(color2)

            # ax2.tick_params(axis='y', color = color2)
            # plt.title("Sub-Basin = " + str(Subid), fontsize = 15)

            # minall = min(min(n1[1]), min(n2[1]))
            # if minall < 0:
            #     minall =0

            # maxall = max(max(n1[1]), max(n2[1]))
            # ax1.set_xlim(minall, maxall)
            #    ax1.set_yticklabels(ax1.get_yticklabels(), color = color1)
            #    ax2.set_yticklabels(ax2.get_yticklabels(), color = color2)

            # # options
            # for key in kwargs.keys():
            # if key == 'legend':
            # ax1.legend(self.FigureOptions['legend'])
            # if key == 'legend_size':
            ax1.legend(
                kwargs["legend"],
                fontsize=int(Visualize.FigureDefaultOptions["legend_size"]),
            )
            # if key == 'xlabel':
            # ax1.set_xlabel(self.FigureOptions['xlabel'])
            # if key == 'ylabel':
            # ax1.set_ylabel(self.FigureOptions['ylabel'])
            # if key == 'labelsize':
            ax1.set_xlabel(
                Visualize.FigureDefaultOptions["xlabel"],
                fontsize=Visualize.FigureDefaultOptions["AxisLabelSize"],
            )
            ax1.set_ylabel(
                Visualize.FigureDefaultOptions["ylabel"],
                fontsize=Visualize.FigureDefaultOptions["AxisLabelSize"],
            )
            # if key == 'fontsize':
            plt.rcParams.update(
                {"font.size": int(Visualize.FigureDefaultOptions["Axisfontsize"])}
            )

            # fig.legend(loc="upper right", bbox_to_anchor=(1,1), bbox_transform=ax1.transAxes,fontsize = 15)

            plt.tight_layout()

        if Save:
            plt.savefig(
                Visualize.FigureDefaultOptions["name"] + "-hist.tif", transparent=True
            )
            # plt.close()

    def ListAttributes(self):
        """ListAttributes.

        Print Attributes List
        """
        print("\n")
        print(
            "Attributes List of: "
            + repr(self.__dict__["name"])
            + " - "
            + self.__class__.__name__
            + " Instance\n"
        )
        self_keys = list(self.__dict__.keys())
        self_keys.sort()
        for key in self_keys:
            if key != "name":
                print(str(key) + " : " + repr(self.__dict__[key]))

        print("\n")


class MidpointNormalize(colors.Normalize):
    """MidpointNormalize.

    !TODO needs docs
    """

    def __init__(self, vmin=None, vmax=None, midpoint=None, clip=False):
        self.midpoint = midpoint
        colors.Normalize.__init__(self, vmin, vmax, clip)

    def __call__(self, value, clip=None):
        """MidpointNormalize.

        ! TODO needs docs

        Parameters
        ----------
        value : TYPE
            DESCRIPTION.
        clip : TYPE, optional
            DESCRIPTION. The default is None.

        Returns
        -------
        TYPE
            DESCRIPTION.
        """
        # I'm ignoring masked values and all kinds of edge cases to make a
        # simple example...
        x, y = [self.vmin, self.midpoint, self.vmax], [0, 0.5, 1]

        return np.ma.masked_array(np.interp(value, x, y))



"""Created on Wed Mar 31 02:10:49 2021.

@author: mofarrag
"""

__name__ = "catchment"

import datetime as dt
import math
import os
from types import ModuleType
from typing import Optional, Union

import geopandas as gpd
import matplotlib.dates as dates
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import statista.metrics as PC
from digitalearth.static import Map
from loguru import logger
from osgeo import gdal
from pyramids.catchment import Catchment as GC
from pyramids.raster import Raster

from Hapi.plot.visualizer import Visualize as Vis


class Catchment:
    """Catchment. The class include methods to read the meteorological and Spatial inputs of the distributed hydrological model. Catchment class also reads the data of the gauges, it is a super class that has the run subclass, so you need to build the catchment object and hand it as an inpit to the Run class to run the model.

    methods:
        1-readRainfall
        2-readTemperature
        3-readET
        4-readFlowAcc
        5-readFlowDir
        6-ReadFlowPathLength
        7-readParameters
        8-readLumpedModel
        9-readLumpedInputs
        10-readGaugeTable
        11-readDischargeGauges
        12-readParametersBounds
        13-extractDischarge
        14-plotHydrograph
        15-PlotDistributedQ
        16-saveResults
    """

    def __init__(
        self,
        name: str,
        startdata: str,
        end: str,
        fmt: str = "%Y-%m-%d",
        SpatialResolution: Optional[str] = "Lumped",
        TemporalResolution: Optional[str] = "Daily",
        RouteRiver: Optional[str] = "Muskingum",
    ):
        """Catchment.

        Parameters
        ----------
        name : [str]
            Name of the Catchment.
        startdata : [str]
            starting date.
        end : [str]
            end date.
        fmt : [str], optional
            format of the given date. The default is "%Y-%m-%d".
        SpatialResolution : TYPE, optional
            Lumped or 'Distributed' . The default is 'Lumped'.
        TemporalResolution : TYPE, optional
            "Hourly" or "Daily". The default is "Daily".

        Returns
        -------
        None.
        """
        self.name = name
        self.start = dt.datetime.strptime(startdata, fmt)
        self.end = dt.datetime.strptime(end, fmt)

        if not SpatialResolution.lower() in ["lumped", "distributed"]:
            raise ValueError(
                "available spatial resolutions are 'lumped' and 'distributed'"
            )
        self.SpatialResolution = SpatialResolution.lower()

        if not TemporalResolution.lower() in ["daily", "hourly"]:
            raise ValueError("available temporal resolutions are 'daily' and 'hourly'")
        self.TemporalResolution = TemporalResolution.lower()
        # assuming the default dt is 1 day
        conversionfactor = (1000 * 24 * 60 * 60) / (1000**2)
        if TemporalResolution.lower() == "daily":
            self.dt = 1  # 24
            self.conversionfactor = conversionfactor * 1
            self.Index = pd.date_range(self.start, self.end, freq="D")
        elif TemporalResolution.lower() == "hourly":
            self.dt = 1  # 24
            self.conversionfactor = conversionfactor * 1 / 24
            self.Index = pd.date_range(self.start, self.end, freq="H")
        else:
            # TODO calculate the temporal resolution factor
            # q mm , area sq km  (1000**2)/1000/f/24/60/60 = 1/(3.6*f)
            # if daily tfac=24 if hourly tfac=1 if 15 min tfac=0.25
            self.conversionfactor = 24

        self.RouteRiver = RouteRiver
        self.Parameters = None
        self.data = None
        self.Prec = None
        self.TS = None
        self.Temp = None
        self.ET = None
        self.ll_temp = None
        self.QGauges = None
        self.Snow = None
        self.Maxbas = None
        self.LumpedModel = None
        self.CatArea = None
        self.InitialCond = None
        self.q_init = None
        self.GaugesTable = None
        self.UB = None
        self.LB = None
        self.cols = None
        self.rows = None
        self.NoDataValue = None
        self.FlowAccArr = None
        self.no_elem = None
        self.acc_val = None
        self.Outlet = None
        self.CellSize = None
        self.px_area = None
        self.px_tot_area = None
        self.FlowDirArr = None
        self.FDT = None
        self.FPLArr = None
        (
            self.DEM,
            self.BankfullDepth,
            self.RiverWidth,
            self.RiverRoughness,
            self.FloodPlainRoughness,
        ) = (None, None, None, None, None)
        self.qout, self.Qtot = None, None
        self.quz_routed, self.qlz_translated, self.statevariables = None, None, None
        self.anim = None
        self.quz, self.qlz = None, None
        self.Qsim = None
        self.Metrics = None

    def readRainfall(self, Path: str, start: str = "", end: str = "", fmt: str = ""):
        """readRainfall.

        Parameters
        ----------
        Path : [String]
            path to the Folder contains precipitation rasters.
        fmt: [str]
            format of the given date
        start: [str]
            start date if you want to read the input temperature for a specific period only,
            if not given all rasters in the given path will be read.
        end: [str]
            end date if you want to read the input temperature for a specific period only,
            if not given all rasters in the given path will be read.


        Returns
        -------
        Prec : [array attribute]
            array containing the spatial rainfall values
        """
        if self.Prec is None:
            # data type
            assert isinstance(Path, str), "PrecPath input should be string type"
            # check wether the path exists or not
            if not os.path.exists(Path):
                raise FileNotFoundError(f"{Path} you have provided does not exist")
            # check wether the folder has the rasters or not
            assert (
                len(os.listdir(Path)) > 0
            ), f"{Path} folder you have provided is empty"
            # read data
            self.Prec = Raster.readRastersFolder(Path, start=start, end=end, fmt=fmt)
            self.TS = (
                self.Prec.shape[2] + 1
            )  # no of time steps =length of time series +1
            assert isinstance(
                self.Prec, np.ndarray
            ), "array should be of type numpy array"

            logger.debug("Rainfall data are read successfully")

    def readTemperature(
        self,
        Path: str,
        ll_temp: Union[list, np.ndarray] = None,
        start: str = "",
        end: str = "",
        fmt: str = "",
    ):
        """readTemperature.

        Parameters
        ----------
        Path : [String]
            path to the Folder contains temperature rasters.
        fmt: [str]
            format of the given date
        start: [str]
            start date if you want to read the input temperature for a specific period only,
            if not given all rasters in the given path will be read.
        end: [str]
            end date if you want to read the input temperature for a specific period only,
            if not given all rasters in the given path will be read.
        ll_temp: [list/ ndarray]
            long term temperature

        Returns
        -------
        Temp : [array attribute]
            array containing the spatial temperature values
        """
        if self.Temp is None:
            # data type
            assert type(Path) == str, "PrecPath input should be string type"
            # check wether the path exists or not
            assert os.path.exists(Path), Path + " you have provided does not exist"
            # check wether the folder has the rasters or not
            assert len(os.listdir(Path)) > 0, (
                Path + " folder you have provided is empty"
            )
            # read data
            self.Temp = Raster.readRastersFolder(Path, start=start, end=end, fmt=fmt)
            assert type(self.Temp) == np.ndarray, "array should be of type numpy array"

            if ll_temp is None:
                self.ll_temp = np.zeros_like(self.Temp, dtype=np.float32)
                avg = self.Temp.mean(axis=2)
                for i in range(self.Temp.shape[0]):
                    for j in range(self.Temp.shape[1]):
                        self.ll_temp[i, j, :] = avg[i, j]

            logger.debug("Temperature data are read successfully")

    def readET(self, Path: str, start: str = "", end: str = "", fmt: str = ""):
        """readET.

        Parameters
        ----------
        Path : [String]
            path to the Folder contains Evapotranspiration rasters.
        fmt: [str]
            format of the given date
        start: [str]
            start date if you want to read the input temperature for a specific period only,
            if not given all rasters in the given path will be read.
        end: [str]
            end date if you want to read the input temperature for a specific period only,
            if not given all rasters in the given path will be read.

        Returns
        -------
        ET : [array attribute]
            array containing the spatial Evapotranspiration values
        """
        if self.ET is None:
            # data type
            assert type(Path) == str, "PrecPath input should be string type"
            # check wether the path exists or not
            assert os.path.exists(Path), Path + " you have provided does not exist"
            # check wether the folder has the rasters or not
            assert len(os.listdir(Path)) > 0, (
                Path + " folder you have provided is empty"
            )
            # read data
            self.ET = Raster.readRastersFolder(Path, start=start, end=end, fmt=fmt)
            assert type(self.ET) == np.ndarray, "array should be of type numpy array"
            logger.debug("Potential Evapotranspiration data are read successfully")

    def readFlowAcc(self, Path: str):
        """readFlowAcc.

        Parameters
        ----------
        Path : [String]
            path to the Flow Accumulation raster of the catchment
            (it should include the raster name and extension).

        Returns
        -------
        FlowAcc : [array attribute]
            array containing the spatial Evapotranspiration values
        rows: [integer]
            number of rows in the flow acc array
        cols:[integer]
            number of columns in the flow acc array
        NoDataValue:[numeric]
            the NoDataValue
        no_elem : [integer]
            number of cells in the domain
        """
        # data type
        assert isinstance(Path, str), "PrecPath input should be string type"
        # check wether the path exists or not
        assert os.path.exists(Path), Path + " you have provided does not exist"
        # check the extension of the accumulation file
        assert (
            Path[-4:] == ".tif"
        ), "please add the extension at the end of the Flow accumulation raster path input"
        # check wether the path exists or not
        assert os.path.exists(Path), Path + " you have provided does not exist"

        FlowAcc = gdal.Open(Path)
        [self.rows, self.cols] = FlowAcc.ReadAsArray().shape
        # check flow accumulation input raster
        self.NoDataValue = FlowAcc.GetRasterBand(1).GetNoDataValue()
        self.FlowAccArr = FlowAcc.ReadAsArray()

        # check if the flow acc array is integer convert it to float
        if self.FlowAccArr.dtype == "int":
            self.FlowAccArr = self.FlowAccArr.astype(float)

        for i in range(self.rows):
            for j in range(self.cols):
                if math.isclose(self.FlowAccArr[i, j], self.NoDataValue, rel_tol=0.001):
                    self.FlowAccArr[i, j] = np.nan

        self.no_elem = np.size(self.FlowAccArr[:, :]) - np.count_nonzero(
            (self.FlowAccArr[np.isnan(self.FlowAccArr)])
        )
        self.acc_val = [
            int(self.FlowAccArr[i, j])
            for i in range(self.rows)
            for j in range(self.cols)
            if not np.isnan(self.FlowAccArr[i, j])
        ]
        self.acc_val = list(set(self.acc_val))
        self.acc_val.sort()
        acc_val_mx = max(self.acc_val)

        if not (acc_val_mx == self.no_elem or acc_val_mx == self.no_elem - 1):
            message = (
                "flow accumulation raster values are not correct max "
                "value should equal number of cells or number of cells -1 "
                f"Max Value in the Flow Acc raster is {acc_val_mx}"
                f" while No of cells are {self.no_elem}"
            )
            logger.debug(message)

        # assert acc_val_mx == self.no_elem or acc_val_mx == self.no_elem -1,

        # location of the outlet
        # outlet is the cell that has the max flow_acc
        self.Outlet = np.where(self.FlowAccArr == np.nanmax(self.FlowAccArr))

        # calculate area covered by cells
        geo_trans = (
            FlowAcc.GetGeoTransform()
        )  # get the coordinates of the top left corner and cell size [x,dx,y,dy]
        dx = np.abs(geo_trans[1]) / 1000.0  # dx in Km
        dy = np.abs(geo_trans[-1]) / 1000.0  # dy in Km
        self.CellSize = dx * 1000

        # area of the cell
        self.px_area = dx * dy
        # no_cells=np.size(raster[:,:])-np.count_nonzero(raster[raster==no_val])
        self.px_tot_area = self.no_elem * self.px_area  # total area of pixels

        logger.debug("Flow Accmulation input is read successfully")

    def readFlowDir(self, Path: str):
        """readFlowDir method reads the flow direction raster.

        Parameters
        ----------
        Path : [str]
            Path to the flow direction raster.

        Returns
        -------
        FlowDirArr : [array].
            array of the flow direction raster
        FDT : [dictionary]
            flow direction table
        """
        # data type
        assert isinstance(Path, str), "PrecPath input should be string type"
        # check wether the path exists or not
        assert os.path.exists(Path), Path + " you have provided does not exist"
        # check the extension of the accumulation file
        assert (
            Path[-4:] == ".tif"
        ), "please add the extension at the end of the Flow accumulation raster path input"
        # check wether the path exists or not
        assert os.path.exists(Path), Path + " you have provided does not exist"
        FlowDir = gdal.Open(Path)

        [rows, cols] = FlowDir.ReadAsArray().shape
        self.FlowDirArr = FlowDir.ReadAsArray().astype(float)
        # check flow direction input raster
        fd_noval = FlowDir.GetRasterBand(1).GetNoDataValue()

        for i in range(rows):
            for j in range(cols):
                if math.isclose(self.FlowDirArr[i, j], fd_noval, rel_tol=0.001):
                    self.FlowDirArr[i, j] = np.nan

        fd_val = [
            int(self.FlowDirArr[i, j])
            for i in range(rows)
            for j in range(cols)
            if not np.isnan(self.FlowDirArr[i, j])
        ]
        fd_val = list(set(fd_val))
        fd_should = [1, 2, 4, 8, 16, 32, 64, 128]
        assert all(
            fd_val[i] in fd_should for i in range(len(fd_val))
        ), "flow direction raster should contain values 1,2,4,8,16,32,64,128 only "

        # create the flow direction table
        self.FDT = GC.flowDirectionTable(FlowDir)
        logger.debug("Flow Direction input is read successfully")

    def ReadFlowPathLength(self, Path: str):
        """ReadFlowPathLength method reads the flow path length.

        Parameters
        ----------
        Path : [str]
            Path to the file.

        Returns
        -------
        FPLArr : [array]
            flpw path length array
        rows : [integer]
            number of rows in the flow acc array
        cols : [integer]
            number of columns in the flow acc array
        NoDataValue : [numeric]
            the NoDataValue
        no_elem : [integer]
            number of cells in the domain
        """
        # data type
        assert isinstance(Path, str), "PrecPath input should be string type"
        # input values
        FPL_ext = Path[-4:]
        assert (
            FPL_ext == ".tif"
        ), "please add the extension at the end of the Flow accumulation raster path input"
        # check wether the path exists or not
        assert os.path.exists(Path), Path + " you have provided does not exist"

        FPL = gdal.Open(Path)
        [self.rows, self.cols] = FPL.ReadAsArray().shape
        self.FPLArr = FPL.ReadAsArray()
        self.NoDataValue = FPL.GetRasterBand(1).GetNoDataValue()

        for i in range(self.rows):
            for j in range(self.cols):
                if math.isclose(self.FPLArr[i, j], self.NoDataValue, rel_tol=0.001):
                    self.FPLArr[i, j] = np.nan
        # check flow accumulation input raster
        self.no_elem = np.size(self.FPLArr[:, :]) - np.count_nonzero(
            (self.FPLArr[np.isnan(self.FPLArr)])
        )

        logger.debug("Flow Path length input is read successfully")

    def ReadRiverGeometry(
        self,
        DEMF: str,
        BankfulldepthF: str,
        RiverWidthF: str,
        RiverRoughnessF: str,
        FloodPlainRoughnessF: str,
    ):
        """ReadRiverGeometry.

        Parameters
        ----------
        DEMF
        BankfulldepthF
        RiverWidthF
        RiverRoughnessF
        FloodPlainRoughnessF

        Returns
        -------
        None
        """
        DEM = gdal.Open(DEMF)
        self.DEM = DEM.ReadAsArray()

        BankfullDepth = gdal.Open(BankfulldepthF)
        self.BankfullDepth = BankfullDepth.ReadAsArray()

        RiverWidth = gdal.Open(RiverWidthF)
        self.RiverWidth = RiverWidth.ReadAsArray()

        RiverRoughness = gdal.Open(RiverRoughnessF)
        self.RiverRoughness = RiverRoughness.ReadAsArray()

        FloodPlainRoughness = gdal.Open(FloodPlainRoughnessF)
        self.FloodPlainRoughness = FloodPlainRoughness.ReadAsArray()

    def readParameters(self, Path: str, Snow: bool = False, Maxbas: bool = False):
        """readParameters.

        readParameters method reads the parameters' raster

        Parameters
        ----------
        Path : [str]
            path to the folder where the raster exist.
        Snow : [integer]
            False if you dont want to run the snow related processes and 1 if there is snow.
            in case of 1 (simulate snow processes) parameters related to snow simulation
            has to be provided. The default is 0.
        Maxbas : [bool], optional
            True if the routing is Maxbas. The default is False.

        Returns
        -------
        Parameters : [array].
            3d array containing the parameters
        Snow : [integer]
            0/1
        Maxbas : [bool]
            True/False
        """
        if self.SpatialResolution.lower() == "distributed":
            # data type
            assert type(Path) == str, "PrecPath input should be string type"
            # check wether the path exists or not
            assert os.path.exists(Path), Path + " you have provided does not exist"
            # check wether the folder has the rasters or not
            assert len(os.listdir(Path)) > 0, (
                Path + " folder you have provided is empty"
            )
            # parameters
            self.Parameters = Raster.readRastersFolder(Path)
        else:
            if not os.path.exists(Path):
                raise FileNotFoundError(
                    "The parameters file you have entered does not exist"
                )

            self.Parameters = pd.read_csv(Path, index_col=0, header=None)[1].tolist()

        assert (
            not Snow or Snow
        ), " snow input defines whether to consider snow subroutine or not it has to be True or False"

        self.Snow = Snow
        self.Maxbas = Maxbas

        if self.SpatialResolution == "distributed":
            if Snow and Maxbas:
                assert self.Parameters.shape[2] == 16, (
                    "current version of HBV (with snow) takes 16 parameter you have entered "
                    + str(self.Parameters.shape[2])
                )
            elif not Snow and Maxbas:
                assert self.Parameters.shape[2] == 11, (
                    "current version of HBV (with snow) takes 11 parameter you have entered "
                    + str(self.Parameters.shape[2])
                )
            elif Snow and not Maxbas:
                assert self.Parameters.shape[2] == 17, (
                    "current version of HBV (with snow) takes 17 parameter you have entered "
                    + str(self.Parameters.shape[2])
                )
            elif not Snow and not Maxbas:
                assert self.Parameters.shape[2] == 12, (
                    "current version of HBV (with snow) takes 12 parameter you have entered "
                    + str(self.Parameters.shape[2])
                )
        else:
            if Snow and Maxbas:
                assert len(self.Parameters) == 16, (
                    "current version of HBV (with snow) takes 16 parameter you have entered "
                    + str(len(self.Parameters))
                )
            elif not Snow and Maxbas:
                if len(self.Parameters) != 11:
                    raise ValueError(
                        f"current version of HBV (with snow) takes 11 parameter you have entered {len(self.Parameters)}"
                    )

            elif Snow and not Maxbas:
                assert len(self.Parameters) == 17, (
                    "current version of HBV (with snow) takes 17 parameter you have entered "
                    + str(len(self.Parameters))
                )
            elif not Snow and not Maxbas:
                assert len(self.Parameters) == 12, (
                    "current version of HBV (with snow) takes 12 parameter you have entered "
                    + str(len(self.Parameters))
                )

        logger.debug("Parameters are read successfully")

    def readLumpedModel(
        self, LumpedModel, CatArea: Union[float, int], InitialCond: list, q_init=None
    ):
        """readLumpedModel.

        Parameters
        ----------
        LumpedModel : [module]
            HBV module.
        CatArea : [numeric]
            Catchment area (km2).
        InitialCond : [list]
            list of the inial condition [SnowPack,SoilMoisture,Upper Zone,
                                         Lower Zone, Water Content].
        q_init : [numeric], optional
            initial discharge. The default is None.

        Returns
        -------
        LumpedModel : [module].
            the lumped conceptual model.
        q_init : [numeric]
            initial discharge.
        InitialCond : [list]
            initial conditions.
        """

        assert isinstance(
            LumpedModel, ModuleType
        ), "ConceptualModel should be a module or a python file contains functions "

        self.LumpedModel = LumpedModel
        self.CatArea = CatArea

        assert (
            len(InitialCond) == 5
        ), f"state variables are 5 and the given initial values are {len(InitialCond)}"

        self.InitialCond = InitialCond

        if q_init is not None:
            assert type(q_init) == float, "q_init should be of type float"
        self.q_init = q_init

        if self.InitialCond is not None:
            assert isinstance(self.InitialCond, list), "init_st should be of type list"

        logger.debug("Lumped model is read successfully")

    def readLumpedInputs(self, Path: str, ll_temp: Union[list, np.ndarray] = None):
        """readLumpedInputs. readLumpedInputs method read the meteorological data of lumped mode.

        [precipitation, Evapotranspiration, temperature, long term average temperature]

        Parameters
        ----------
        Path : [string]
            Path to the input file, data has to be in the order of
            [date, precipitation, ET, Temp].
        ll_temp : [bool], optional
            average long term temperature, if None it is calculated inside the
            code. The default is None.

        Returns
        -------
        data : [array].
            meteorological data.
        ll_temp : [array]
            average long term temperature.
        """
        self.data = pd.read_csv(
            Path, header=0, delimiter=",", index_col=0  # "\t", #skiprows=11,
        )
        self.data = self.data.values

        if ll_temp is None:
            self.ll_temp = np.zeros(shape=(len(self.data)), dtype=np.float32)
            self.ll_temp = self.data[:, 2].mean()

        assert (
            np.shape(self.data)[1] == 3 or np.shape(self.data)[1] == 4
        ), " meteorological data should be of length at least 3 (prec, ET, temp) or 4(prec, ET, temp, tm) "

        logger.debug("Lumped Model inputs are read successfully")

    def readGaugeTable(self, Path: str, FlowaccPath: str = "", fmt: str = "%Y-%m-%d"):
        """readGaugeTable. readGaugeTable reads the table where the data about the gauges are listed.

        [x coordinate, y coordinate, 'area ratio', 'weight'], the coordinates are
        mandatory to enter , to locate the gauges and be able to extract the
        discharge at the coresponding cells.

        Parameters
        ----------
        Path : [str]
            Path to the gauge file.
        FlowaccPath : [str], optional
            Path to the Flow acc raster. The default is ''.
        fmt: [str]
            Default is "%Y-%m-%d"

        Returns
        -------
        hm_gauges : [dataframe]
            the table of the gauges.
        """
        # read the gauge table
        if Path.endswith(".geojson"):
            self.GaugesTable = gpd.read_file(Path, driver="GeoJSON")
        else:
            self.GaugesTable = pd.read_csv(Path)
        col_list = self.GaugesTable.columns.tolist()

        if "start" in col_list:
            for i in range(len(self.GaugesTable)):
                self.GaugesTable.loc[i, "start"] = dt.datetime.strptime(
                    self.GaugesTable.loc[i, "start"], fmt
                )
                self.GaugesTable.loc[i, "end"] = dt.datetime.strptime(
                    self.GaugesTable.loc[i, "end"], fmt
                )
        if FlowaccPath != "" and "cell_row" not in col_list:
            # if hasattr(self, 'FlowAcc'):
            FlowAcc = gdal.Open(FlowaccPath)
            # calculate the nearest cell to each station
            self.GaugesTable.loc[:, ["cell_row", "cell_col"]] = GC.nearestCell(
                FlowAcc, self.GaugesTable[["id", "x", "y"]][:]
            )  # ,'weight'

        logger.debug("Gauge Table is read successfully")

    def readDischargeGauges(
        self,
        Path: str,
        delimiter: str = ",",
        column: str = "id",
        fmt: str = "%Y-%m-%d",
        Split: bool = False,
        Date1: str = "",
        Date2: str = "",
        readfrom: str = "",
    ):
        """readDischargeGauges. readDischargeGauges method read the gauge discharge data, discharge of each gauge has to be stored separetly in a file, and the name of the file has to be stored in the Gauges table you entered ubing the method "readGaugeTable" under the column "id", the file should contains the date at the first column.

        Parameters
        ----------
        Path : [str]
            Path the the gauge discharge data.
        delimiter : [str], optional
            the delimiter between the date and the discharge column. The default is ",".
        column : [str], optional
            the name of the column where you stored the files names. The default is 'id'.
        fmt : [str], optional
            date format. The default is "%Y-%m-%d".
        Split : bool, optional
            True if you want to split the data between two dates. The default is False.
        Date1 : [str], optional
            start date. The default is ''.
        Date2 : [str], optional
            end date. The default is ''.
        readfrom: [str]
            Default is "".

        Returns
        -------
        GaugesTable : [dataframe].
            dataframe containing the discharge data
        """
        if self.TemporalResolution.lower() == "daily":
            ind = pd.date_range(self.start, self.end, freq="D")
        else:
            ind = pd.date_range(self.start, self.end, freq="H")

        if self.SpatialResolution.lower() == "distributed":
            assert hasattr(self, "GaugesTable"), "please read the gauges table first"

            self.QGauges = pd.DataFrame(
                index=ind, columns=self.GaugesTable[column].tolist()
            )

            for i in range(len(self.GaugesTable)):
                name = self.GaugesTable.loc[i, "id"]
                if readfrom != "":
                    f = pd.read_csv(
                        Path + str(name) + ".csv",
                        index_col=0,
                        delimiter=delimiter,
                        skiprows=readfrom,
                    )  # ,#delimiter="\t"
                else:
                    f = pd.read_csv(
                        Path + str(name) + ".csv",
                        header=0,
                        index_col=0,
                        delimiter=delimiter,
                    )

                f.index = [dt.datetime.strptime(i, fmt) for i in f.index.tolist()]

                self.QGauges[int(name)] = f.loc[self.start : self.end, f.columns[-1]]
        else:
            if not os.path.exists(Path):
                raise FileNotFoundError(
                    f"The file you have entered{Path} does not exist"
                )

            self.QGauges = pd.DataFrame(index=ind)
            f = pd.read_csv(
                Path, header=0, index_col=0, delimiter=delimiter
            )  # ,#delimiter="\t", skiprows=11,
            f.index = [dt.datetime.strptime(i, fmt) for i in f.index.tolist()]
            self.QGauges[f.columns[0]] = f.loc[self.start : self.end, f.columns[0]]

        if Split:
            Date1 = dt.datetime.strptime(Date1, fmt)
            Date2 = dt.datetime.strptime(Date2, fmt)
            self.QGauges = self.QGauges.loc[Date1:Date2]

        logger.debug("Gauges data are read successfully")

    def readParametersBounds(
        self,
        UB: Union[list, np.ndarray],
        LB: Union[list, np.ndarray],
        Snow: bool = False,
        Maxbas: bool = False,
    ):
        """readParametersBounds. readParametersBounds method reads the lower and upper boundarys for each parameter.

        Parameters
        ----------
        UB : [list]
            upper bound.
        LB : [list]
            lower bound.
        Snow : [integer]
            0 if you dont want to run the snow related processes and 1 if there is snow.
            in case of 1 (simulate snow processes) parameters related to snow simulation
            has to be provided. The default is 0.
        Maxbas: [bool]
            True if the parameters has maxbas.

        Returns
        -------
        UB : [list]
            upper bound.
        LB : [list]
            lower bound.
        Snow : [integer]
            Snow
        """
        assert len(UB) == len(LB), "length of UB should be the same like LB"
        self.UB = np.array(UB)
        self.LB = np.array(LB)

        if not isinstance(Snow, bool):
            raise ValueError(
                " snow input defines whether to consider snow subroutine or not it has to be "
                "True or False"
            )
        self.Snow = Snow
        self.Maxbas = Maxbas

        logger.debug("Parameters bounds are read successfully")

    def extractDischarge(
        self, CalculateMetrics=True, FW1=False, Factor=None, OnlyOutlet=False
    ):
        """extractDischarge. extractDischarge method extracts and sums the discharge from the Quz_routed and Qlz_translated arrays at the location of the gauges.

        Parameters
        ----------
        CalculateMetrics : bool, optional
            True if you want to calculate the performance metrics. The default is True.
        FW1 : [bool], optional
            True if the routing function is Maxbas. The default is False.
        Factor : [list/None]
            list of factor if you want to multiply the simulated discharge by
            a factor you have to provide a list of the factor (as many factors
            as the number of gauges). The default is False.
        OnlyOutlet: [bool]
            True to extract the discharge only at the outlet.

        Returns
        -------
        Qsim : [dataframe]
            dataframe containing the discharge time series of the cells where
            the gauges are located.
        Metrics : [dataframe]
            data frame containing the followinf metrics ['RMSE', 'NSE',
            'NSEhf', 'KGE', 'WB','Pearson-CC','R2'] calculated between the simulated
            hydrographs and the gauge data
        """

        if not FW1:
            self.Qsim = pd.DataFrame(index=self.Index, columns=self.QGauges.columns)
            if CalculateMetrics:
                index = ["RMSE", "NSE", "NSEhf", "KGE", "WB", "Pearson-CC", "R2"]
                self.Metrics = pd.DataFrame(index=index, columns=self.QGauges.columns)
            # sum the lower zone and the upper zone discharge
            outletx = self.Outlet[0][0]
            outlety = self.Outlet[1][0]

            # self.qout = self.qlz_translated[outletx,outlety,:] + self.quz_routed[outletx,outlety,:]
            # self.Qtot = self.qlz_translated + self.quz_routed
            self.qout = self.Qtot[outletx, outlety, :]

            for i in range(len(self.GaugesTable)):
                Xind = int(self.GaugesTable.loc[self.GaugesTable.index[i], "cell_row"])
                Yind = int(self.GaugesTable.loc[self.GaugesTable.index[i], "cell_col"])
                gaugeid = self.GaugesTable.loc[self.GaugesTable.index[i], "id"]

                # Quz = np.reshape(self.quz_routed[Xind,Yind,:-1],self.TS-1)
                # Qlz = np.reshape(self.qlz_translated[Xind,Yind,:-1],self.TS-1)
                # Qsim = Quz + Qlz

                Qsim = np.reshape(self.Qtot[Xind, Yind, :-1], self.TS - 1)
                if Factor is not None:
                    self.Qsim.loc[:, gaugeid] = Qsim * Factor[i]
                else:
                    self.Qsim.loc[:, gaugeid] = Qsim

                if CalculateMetrics:
                    Qobs = self.QGauges.loc[:, gaugeid]
                    self.Metrics.loc["RMSE", gaugeid] = round(PC.RMSE(Qobs, Qsim), 3)
                    self.Metrics.loc["NSE", gaugeid] = round(PC.NSE(Qobs, Qsim), 3)
                    self.Metrics.loc["NSEhf", gaugeid] = round(PC.NSEHF(Qobs, Qsim), 3)
                    self.Metrics.loc["KGE", gaugeid] = round(PC.KGE(Qobs, Qsim), 3)
                    self.Metrics.loc["WB", gaugeid] = round(PC.WB(Qobs, Qsim), 3)
                    self.Metrics.loc["Pearson-CC", gaugeid] = round(
                        PC.PearsonCorre(Qobs, Qsim), 3
                    )
                    self.Metrics.loc["R2", gaugeid] = round(PC.R2(Qobs, Qsim), 3)
        elif FW1 or OnlyOutlet:
            self.Qsim = pd.DataFrame(index=self.Index)
            gaugeid = self.GaugesTable.loc[self.GaugesTable.index[-1], "id"]
            Qsim = np.reshape(self.qout, self.TS - 1)
            self.Qsim.loc[:, gaugeid] = Qsim

            if CalculateMetrics:
                index = ["RMSE", "NSE", "NSEhf", "KGE", "WB", "Pearson-CC", "R2"]
                self.Metrics = pd.DataFrame(index=index)

                # if CalculateMetrics:
                Qobs = self.QGauges.loc[:, gaugeid]
                self.Metrics.loc["RMSE", gaugeid] = round(PC.RMSE(Qobs, Qsim), 3)
                self.Metrics.loc["NSE", gaugeid] = round(PC.NSE(Qobs, Qsim), 3)
                self.Metrics.loc["NSEhf", gaugeid] = round(PC.NSEHF(Qobs, Qsim), 3)
                self.Metrics.loc["KGE", gaugeid] = round(PC.KGE(Qobs, Qsim), 3)
                self.Metrics.loc["WB", gaugeid] = round(PC.WB(Qobs, Qsim), 3)
                self.Metrics.loc["Pearson-CC", gaugeid] = round(
                    PC.PearsonCorre(Qobs, Qsim), 3
                )
                self.Metrics.loc["R2", gaugeid] = round(PC.R2(Qobs, Qsim), 3)

    def plotHydrograph(
        self,
        plotstart: str,
        plotend: str,
        gaugei: int,
        Hapicolor: Union[tuple, str] = "#004c99",
        gaugecolor: Union[tuple, str] = "#DC143C",
        linewidth: int = 3,
        Hapiorder: int = 1,
        Gaugeorder: int = 0,
        labelfontsize: int = 10,
        XMajorfmt: str = "%Y-%m-%d",
        Noxticks: int = 5,
        Title: str = "",
        Xaxis_fmt: str = "%d\n%m",
        label: str = "",
        fmt: str = "%Y-%m-%d",
    ):
        """plotHydrograph. plot the simulated and gauge hydrograph.

        Parameters
        ----------
        plotstart : [str]
            starting date.
        plotend : [str]
            end date.
        gaugei : [integer]
            order if the gauge in the GaugeTable.
        Hapicolor : [str], optional
            color of the Simulated hydrograph. The default is "#004c99".
        gaugecolor : [str], optional
            color of the gauge. The default is "#DC143C".
        linewidth : [numeric], optional
            line width. The default is 3.
        Hapiorder : [integer], optional
            the order of the simulated hydrograph to control which hydrograph
            is in the front . The default is 1.
        Gaugeorder : TYPE, optional
            the order of the simulated hydrograph to control which hydrograph
            is in the front . The default is 0.
        labelfontsize : numeric, optional
            label size. The default is 10.
        XMajorfmt : [str], optional
            format of the x-axis labels. The default is '%Y-%m-%d'.
        Noxticks : [integer], optional
            number of x-axis ticks. The default is 5.
        Title: [str]
            Default is "".
        Xaxis_fmt: [str]
            Default is "%d\n%m".
        label: [str]
            Default is "".
        fmt:[str]
            Default is "%Y-%m-%d".

        Returns
        -------
        fig : TYPE
            DESCRIPTION.
        ax : [matplotlib axes]
            you can control the figure from the axes.
        """
        plotstart = dt.datetime.strptime(plotstart, fmt)
        plotend = dt.datetime.strptime(plotend, fmt)

        fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(6, 5))

        if self.SpatialResolution == "distributed":
            gaugeid = self.GaugesTable.loc[gaugei, "id"]

            if Title == "":
                Title = "Gauge - " + str(self.GaugesTable.loc[gaugei, "name"])

            if label == "":
                label = str(self.GaugesTable.loc[gaugei, "name"])

            ax.plot(
                self.Qsim.loc[plotstart:plotend, gaugeid],
                "-.",
                label=label,
                linewidth=linewidth,
                color=Hapicolor,
                zorder=Hapiorder,
            )
            ax.set_title(Title, fontsize=20)
        else:
            gaugeid = self.QGauges.columns[0]
            if Title == "":
                Title = "Gauge - " + str(gaugeid)
            if label == "":
                label = str(gaugeid)

            ax.plot(
                self.Qsim.loc[plotstart:plotend, gaugeid],
                "-.",
                label=Title,
                linewidth=linewidth,
                color=Hapicolor,
                zorder=Hapiorder,
            )
            ax.set_title(Title, fontsize=20)

        ax.plot(
            self.QGauges.loc[plotstart:plotend, gaugeid],
            label="Gauge",
            linewidth=linewidth,
            color=gaugecolor,
            zorder=Gaugeorder,
        )

        ax.tick_params(axis="both", which="major", labelsize=labelfontsize)
        # ax.locator_params(axis="x", nbins=4)

        XMajorfmt = dates.DateFormatter(XMajorfmt)
        ax.xaxis.set_major_formatter(XMajorfmt)
        # ax.xaxis.set_minor_locator(dates.WeekdayLocator(byweekday=(1),
        # interval=1))

        ax.xaxis.set_minor_formatter(dates.DateFormatter(Xaxis_fmt))

        ax.xaxis.set_major_locator(plt.MaxNLocator(Noxticks))

        ax.legend(fontsize=12)
        ax.set_xlabel("Time", fontsize=12)
        ax.set_ylabel("Discharge m3/s", fontsize=12)
        plt.tight_layout()

        if self.Metrics:
            logger.debug("----------------------------------")
            logger.debug("Gauge - " + str(gaugeid))
            logger.debug("RMSE= " + str(round(self.Metrics.loc["RMSE", gaugeid], 2)))
            logger.debug("NSE= " + str(round(self.Metrics.loc["NSE", gaugeid], 2)))
            logger.debug("NSEhf= " + str(round(self.Metrics.loc["NSEhf", gaugeid], 2)))
            logger.debug("KGE= " + str(round(self.Metrics.loc["KGE", gaugeid], 2)))
            logger.debug("WB= " + str(round(self.Metrics.loc["WB", gaugeid], 2)))
            logger.debug(
                "Pearson-CC= " + str(round(self.Metrics.loc["Pearson-CC", gaugeid], 2))
            )
            logger.debug("R2= " + str(round(self.Metrics.loc["R2", gaugeid], 2)))

        return fig, ax

    def plotDistributedResults(
        self,
        start: str,
        end: str,
        fmt: str = "%Y-%m-%d",
        Option: int = 1,
        Gauges: bool = False,
        **kwargs,
    ):
        """plotDistributedResults. plotDistributedResults animate the time series of the meteorological inputs and the result calculated by the model  like the total discharge, upper zone, and lower zone discharge and the state variables.

        Parameters
        ----------
        start : [str]
            starting date
        end : [str]
            end date
        fmt : [str]
            format of the gicen date. The default is "%Y-%m-%d"
        Option : [str]
            1- Total discharge, 2-Upper zone discharge, 3-ground water,
            4-Snowpack state variable, 5-Soil moisture, 6-Upper zone,
            7-Lower zone, 8-Water content, 9-Precipitation input. 10-ET,
            11-Temperature. The default is 1
        Gauges : [str]
            . The default is False
        **kwargs :
            possible keyward args
            TicksSpacing : [integer], optional
                Spacing in the colorbar ticks. The default is 2.
            Figsize : [tuple], optional
                figure size. The default is (8,8).
            PlotNumbers : [bool], optional
                True to plot the values intop of each cell. The default is True.
            NumSize : integer, optional
                size of the numbers plotted intop of each cells. The default is 8.
            Title : [str], optional
                title of the plot. The default is 'Total Discharge'.
            titlesize : [integer], optional
                title size. The default is 15.
            Backgroundcolorthreshold : [float/integer], optional
                threshold value if the value of the cell is greater, the plotted
                numbers will be black and if smaller the plotted number will be white
                if None given the maxvalue/2 will be considered. The default is None.
            textcolors : TYPE, optional
                Two colors to be used to plot the values i top of each cell. The default is ("white","black").
            cbarlabel : str, optional
                label of the color bar. The default is 'Discharge m3/s'.
            cbarlabelsize : integer, optional
                size of the color bar label. The default is 12.
            Cbarlength : [float], optional
                ratio to control the height of the colorbar. The default is 0.75.
            Interval : [integer], optional
                number to controlthe speed of the animation. The default is 200.
            cmap : [str], optional
                color style. The default is 'coolwarm_r'.
            Textloc : [list], optional
                location of the date text. The default is [0.1,0.2].
            Gaugecolor : [str], optional
                color of the points. The default is 'red'.
            Gaugesize : [integer], optional
                size of the points. The default is 100.
            ColorScale : integer, optional
                there are 5 options to change the scale of the colors. The default is 1.
                1- ColorScale 1 is the normal scale
                2- ColorScale 2 is the power scale
                3- ColorScale 3 is the SymLogNorm scale
                4- ColorScale 4 is the PowerNorm scale
                5- ColorScale 5 is the BoundaryNorm scale
                ------------------------------------------------------------------
                gamma : [float], optional
                    value needed for option 2 . The default is 1./2..
                linthresh : [float], optional
                    value needed for option 3. The default is 0.0001.
                linscale : [float], optional
                    value needed for option 3. The default is 0.001.
                midpoint : [float], optional
                    value needed for option 5. The default is 0.
                ------------------------------------------------------------------
            orientation : [string], optional
                orintation of the colorbar horizontal/vertical. The default is 'vertical'.
            rotation : [number], optional
                rotation of the colorbar label. The default is -90.
            **kwargs : [dict]
                keys:
                    Points : [dataframe].
                        dataframe contains two columns 'cell_row', and cell_col to
                        plot the point at this location

        Returns
        -------
        animation.FuncAnimation.
        """
        start = dt.datetime.strptime(start, fmt)
        end = dt.datetime.strptime(end, fmt)

        starti = np.where(self.Index == start)[0][0]
        endi = np.where(self.Index == end)[0][0]

        if 1 > Option > 11:
            raise ValueError("Plotting options are from 1 to 11")

        if Option == 1:
            self.Qtot[self.FlowAccArr == self.NoDataValue, :] = np.nan
            Arr = self.Qtot[:, :, starti:endi]
            Title = "Total Discharge"
        elif Option == 2:
            self.quz_routed[self.FlowAccArr == self.NoDataValue, :] = np.nan
            Arr = self.quz_routed[:, :, starti:endi]
            Title = "Surface Flow"
        elif Option == 3:
            self.qlz_translated[self.FlowAccArr == self.NoDataValue, :] = np.nan
            Arr = self.qlz_translated[:, :, starti:endi]
            Title = "Ground Water Flow"
        elif Option == 4:
            self.statevariables[self.FlowAccArr == self.NoDataValue, :, 0] = np.nan
            Arr = self.statevariables[:, :, starti:endi, 0]
            Title = "Snow Pack"
        elif Option == 5:
            self.statevariables[self.FlowAccArr == self.NoDataValue, :, 1] = np.nan
            Arr = self.statevariables[:, :, starti:endi, 1]
            Title = "Soil Moisture"
        elif Option == 6:
            self.statevariables[self.FlowAccArr == self.NoDataValue, :, 2] = np.nan
            Arr = self.statevariables[:, :, starti:endi, 2]
            Title = "Upper Zone"
        elif Option == 7:
            self.statevariables[self.FlowAccArr == self.NoDataValue, :, 3] = np.nan
            Arr = self.statevariables[:, :, starti:endi, 3]
            Title = "Lower Zone"
        elif Option == 8:
            self.statevariables[self.FlowAccArr == self.NoDataValue, :, 4] = np.nan
            Arr = self.statevariables[:, :, starti:endi, 4]
            Title = "Water Content"
        elif Option == 9:
            self.Prec[self.FlowAccArr == self.NoDataValue, :] = np.nan
            Arr = self.Prec[:, :, starti:endi]
            Title = "Precipitation"
        elif Option == 10:
            self.ET[self.FlowAccArr == self.NoDataValue, :] = np.nan
            Arr = self.ET[:, :, starti:endi]
            Title = "ET"
        elif Option == 11:
            self.Temp[self.FlowAccArr == self.NoDataValue, :] = np.nan
            Arr = self.Temp[:, :, starti:endi]
            Title = "Temperature"

        Time = self.Index[starti:endi]

        if Gauges:
            kwargs["Points"] = self.GaugesTable

        anim = Map.AnimateArray(Arr, Time, self.no_elem, Title=Title, **kwargs)

        self.anim = anim

        return anim

    def saveAnimation(self, VideoFormat="gif", Path="", SaveFrames=20):
        """saveAnimation. saveAnimation.

        Parameters
        ----------
        VideoFormat : [str], optional
            possible formats ['mp4','mov', 'avi', 'gif']. The default is "gif".
        Path : [str], optional
            path inclusinf the video format. The default is ''.
        SaveFrames : [integer], optional
            speed of the video. The default is 20.

        Returns
        -------
        None.
        """
        Vis.SaveAnimation(
            self.anim, VideoFormat=VideoFormat, Path=Path, SaveFrames=SaveFrames
        )

    def saveResults(
        self,
        FlowAccPath: str = "",
        Result: int = 1,
        start: str = "",
        end: str = "",
        Path: str = "",
        Prefix: str = "",
        fmt: str = "%Y-%m-%d",
    ):
        """saveResults. saveResults save the results into rasters.

        Parameters
        ----------
        FlowAccPath : [path]
            Path to Flow acc raster.
        Result : [integer], optional
            1 for the total discharge, 2 for the upper zone discharge, 3 for
            the lower zone discharge, 4 for the snow pack, 5 for the soil
            moisture, 6 upper zone, 7 for the lower zone, 8 for the water content.
            The default is 1.
        start : [str], optional
            start date. The default is ''.
        end : [str], optional
            end date. The default is ''.
        Path : [str], optional
            PAth to the directory where you want to save the results. The default is ''.
        Prefix : [str], optional
            prefix to add to the name of the result files. The default is ''.
        fmt : [str], optional
            format of the date. The default is "%Y-%m-%d".

        Returns
        -------
        None.
        """
        if start == "":
            start = self.Index[0]
        else:
            start = dt.datetime.strptime(start, fmt)

        if end == "":
            end = self.Index[-1]
        else:
            end = dt.datetime.strptime(end, fmt)

        starti = np.where(self.Index == start)[0][0]
        endi = np.where(self.Index == end)[0][0] + 1

        if self.SpatialResolution == "distributed":
            assert (
                FlowAccPath != ""
            ), "Please enter the  FlowAccPath parameter to the saveResults method"
            src = gdal.Open(FlowAccPath)

            if Prefix == "":
                Prefix = "Result_"

            # create list of names
            Path = Path + Prefix
            names = [Path + str(i)[:10] for i in self.Index[starti:endi]]
            names = [i.replace("-", "_") for i in names]
            names = [i.replace(" ", "_") for i in names]
            names = [i + ".tif" for i in names]
            if Result == 1:
                Raster.rastersLike(src, self.Qtot[:, :, starti:endi], names)
            elif Result == 2:
                Raster.rastersLike(src, self.quz_routed[:, :, starti:endi], names)
            elif Result == 3:
                Raster.rastersLike(src, self.qlz_translated[:, :, starti:endi], names)
            elif Result == 4:
                Raster.rastersLike(
                    src, self.statevariables[:, :, starti:endi, 0], names
                )
            elif Result == 5:
                Raster.rastersLike(
                    src, self.statevariables[:, :, starti:endi, 1], names
                )
            elif Result == 6:
                Raster.rastersLike(
                    src, self.statevariables[:, :, starti:endi, 2], names
                )
            elif Result == 7:
                Raster.rastersLike(
                    src, self.statevariables[:, :, starti:endi, 3], names
                )
            elif Result == 8:
                Raster.rastersLike(
                    src, self.statevariables[:, :, starti:endi, 4], names
                )
        else:
            ind = pd.date_range(start, end, freq="D")
            data = pd.DataFrame(index=ind)

            data["date"] = ["'" + str(i)[:10] + "'" for i in data.index]

            if Result == 1:
                data["Qsim"] = self.Qsim[starti:endi]
                data.to_csv(Path, index=False, float_format="%.3f")
            elif Result == 2:
                data["Quz"] = self.quz[starti:endi]
                data.to_csv(Path, index=False, float_format="%.3f")
            elif Result == 3:
                data["Qlz"] = self.qlz[starti:endi]
                data.to_csv(Path, index=False, float_format="%.3f")
            elif Result == 4:
                data[["SP", "SM", "UZ", "LZ", "WC"]] = self.statevariables[
                    starti:endi, :
                ]
                data.to_csv(Path, index=False, float_format="%.3f")
            elif Result == 5:
                data["Qsim"] = self.Qsim[starti:endi]
                data["Quz"] = self.quz[starti:endi]
                data["Qlz"] = self.qlz[starti:endi]
                data[["SP", "SM", "UZ", "LZ", "WC"]] = self.statevariables[
                    starti:endi, :
                ]
                data.to_csv(Path, index=False, float_format="%.3f")
            else:
                assert False, "the possible options are from 1 to 5"

        logger.debug("Data is saved successfully")

    def ListAttributes(self):
        """Print Attributes List."""

        logger.debug("\n")
        logger.debug(
            "Attributes List of: "
            + repr(self.__dict__["name"])
            + " - "
            + self.__class__.__name__
            + " Instance\n"
        )
        self_keys = list(self.__dict__.keys())
        self_keys.sort()
        for key in self_keys:
            if key != "name":
                logger.debug(str(key) + " : " + repr(self.__dict__[key]))

        logger.debug("\n")


class Lake:
    """Lake. Lake class reads the meteorological inputs, and the module to simulate a lake as a lumped model, using a rating curve, the lake and the upstream sub-catchments are going to be considered as one lumped model than result in a discharge input to the lake, the discharge input is going to change the volume of the water in the lake, and from the volume-outflow curve the outflow can be obtained.

    methods:
        1- readMeteoData
        2- readParameters
        3- readLumpedModel
    """

    def __init__(
        self,
        start="",
        end="",
        fmt="%Y-%m-%d",
        TemporalResolution="Daily",
        Split=False,
    ):
        """Lake. Lake class for lake simulation.

        Parameters
        ----------
        start : [str], optional
            start date. The default is ''.
        end : [str], optional
            end date. The default is ''.
        fmt : [str], optional
            date format. The default is "%Y-%m-%d".
        TemporalResolution : [str], optional
            "Daily" ot "Hourly". The default is "Daily".
        Split : bool, optional
            true if you want to split the date between two dates. The default is False.

        Returns
        -------
        None.
        """

        self.OutflowCell = None
        self.Snow = None
        self.Split = Split
        self.start = dt.datetime.strptime(start, fmt)
        self.end = dt.datetime.strptime(end, fmt)

        if TemporalResolution.lower() == "daily":
            self.Index = pd.date_range(start, end, freq="D")
        elif TemporalResolution.lower() == "hourly":
            self.Index = pd.date_range(start, end, freq="H")
        else:
            assert False, "Error"
        self.MeteoData = None
        self.Parameters = None
        self.LumpedModel, self.CatArea, self.LakeArea, self.InitialCond = (
            None,
            None,
            None,
            None,
        )
        self.StageDischargeCurve = None

    def readMeteoData(self, Path: str, fmt: str):
        """readMeteoData. readMeteoData reads the meteorological data for the lake.

        [rainfall, ET, temperature]

        Parameters
        ----------
        Path : [str]
            path to the meteo data file, containing the data in the follwoing
            order [date, rainfall, ET, temperature].
        fmt : [str]
            date format.

        Returns
        -------
        MeteoData : [array].
            array containing the meteorological data
        """

        df = pd.read_csv(Path, index_col=0)
        df.index = [dt.datetime.strptime(date, fmt) for date in df.index]

        if self.Split:
            df = df.loc[self.start : self.end, :]

        self.MeteoData = df.values  # lakeCalibArray = lakeCalibArray[:,0:-1]

        logger.debug("Lake Meteo data are read successfully")

    def readParameters(self, Path):
        """readParameters. readParameters method reads the lake parameters.

        Parameters
        ----------
        Path : [str]
            Path to the parameter file.

        Returns
        -------
        Parameters : [array].
        """
        Parameters = np.loadtxt(Path).tolist()
        self.Parameters = Parameters
        logger.debug("Lake Parameters are read successfully")

    def readLumpedModel(
        self,
        LumpedModel,
        CatArea,
        LakeArea,
        InitialCond,
        OutflowCell,
        StageDischargeCurve,
        Snow,
    ):
        """readLumpedModel.

        readLumpedModel reads the lumped model module

        Parameters
        ----------
        LumpedModel : [module]
            lumped conceptual model.
        CatArea : [numeric]
            catchment area in mk2.
        LakeArea : [numeric]
            area of the lake in km2.
        InitialCond : [list]
            initial condition [Snow Pack, Soil Moisture, Upper Zone, Lower Zone,
                               Water Content, Lake volume].
        OutflowCell : [list]
            indeces of the cell where the lake hydrograph is going to be added.
        StageDischargeCurve : [array]
            volume-outflow curve.
        Snow : [integer]
            0 if you dont want to run the snow related processes and 1 if there is snow.
            in case of 1 (simulate snow processes) parameters related to snow simulation
            has to be provided. The default is 0.

        Returns
        -------
        LumpedModel : [module]
            HBV module.
        CatArea : [numeric]
            Catchment area (km2).
        InitialCond : [list]
            list of the inial condition [SnowPack,SoilMoisture,Upper Zone,
                                         Lower Zone, Water Content].
        Snow : [integer]
            0/1
        StageDischargeCurve : [array]
        """
        assert isinstance(
            LumpedModel, ModuleType
        ), "ConceptualModel should be a module or a python file contains functions "
        self.LumpedModel = LumpedModel

        self.CatArea = CatArea
        self.LakeArea = LakeArea
        self.InitialCond = InitialCond

        if self.InitialCond is not None:
            assert type(self.InitialCond) == list, "init_st should be of type list"

        self.Snow = Snow
        self.OutflowCell = OutflowCell
        self.StageDischargeCurve = StageDischargeCurve
        logger.debug("Lumped model is read successfully")


if __name__ == "__main__":
    logger.debug("Catchment module")
